# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pathlib import Path

import pytest

import ansys.fluent.core as pyfluent
from ansys.fluent.core import FluentVersion, MeshingEvent, SolverEvent, examples
from ansys.fluent.core.pyfluent_warnings import PyFluentDeprecationWarning


def test_receive_events_on_case_loaded(new_solver_session) -> None:

    def on_case_loaded_old(session_id, event_info):
        on_case_loaded_old.loaded = True

    on_case_loaded_old.loaded = False

    def on_case_loaded_old_with_args(x, y, session_id, event_info):
        on_case_loaded_old_with_args.state = dict(x=x, y=y)

    on_case_loaded_old_with_args.state = None

    def on_case_loaded(session, event_info):
        on_case_loaded.loaded = True
        if session.get_fluent_version() >= FluentVersion.v232:
            assert Path(event_info.case_file_name).name == Path(case_file_name).name
            with pytest.warns(PyFluentDeprecationWarning):
                assert Path(event_info.casefilepath).name == Path(case_file_name).name

    on_case_loaded.loaded = False

    def on_case_loaded_with_args_optional_first(x, y, session, event_info):
        on_case_loaded_with_args_optional_first.state = dict(x=x, y=y)

    on_case_loaded_with_args_optional_first.state = None

    def on_case_loaded_with_args(session, event_info, x, y):
        on_case_loaded_with_args.state = dict(x=x, y=y)

    on_case_loaded_with_args.state = None

    solver = new_solver_session

    solver.events.register_callback(SolverEvent.CASE_LOADED, on_case_loaded_old)

    solver.events.register_callback(
        SolverEvent.CASE_LOADED, on_case_loaded_old_with_args, 12, y=42
    )

    solver.events.register_callback(SolverEvent.CASE_LOADED, on_case_loaded)

    solver.events.register_callback(
        SolverEvent.CASE_LOADED, on_case_loaded_with_args_optional_first, 12, y=42
    )

    solver.events.register_callback(
        SolverEvent.CASE_LOADED, on_case_loaded_with_args, 12, y=42
    )

    case_file_name = examples.download_file(
        "mixing_elbow.cas.h5", "pyfluent/mixing_elbow"
    )

    assert not on_case_loaded_old.loaded
    assert not on_case_loaded.loaded
    assert not on_case_loaded_old_with_args.state
    assert not on_case_loaded_with_args_optional_first.state
    assert not on_case_loaded_with_args.state

    try:
        solver.settings.file.read_case(file_name=case_file_name)
    except AttributeError:
        solver.tui.file.read_case(case_file_name)

    assert on_case_loaded_old.loaded
    assert on_case_loaded.loaded
    assert on_case_loaded_old_with_args.state == dict(x=12, y=42)
    assert on_case_loaded_with_args_optional_first.state == dict(x=12, y=42)
    assert on_case_loaded_with_args.state == dict(x=12, y=42)


def test_receive_meshing_events_on_case_loaded(new_meshing_session) -> None:

    case_file_name = examples.download_file(
        "mixing_elbow.cas.h5", "pyfluent/mixing_elbow"
    )

    def on_case_loaded(session, event_info):
        on_case_loaded.loaded = True

    on_case_loaded.loaded = False

    meshing = new_meshing_session

    meshing.events.register_callback(MeshingEvent.CASE_LOADED, on_case_loaded)

    assert not on_case_loaded.loaded

    meshing.tui.file.read_case(case_file_name)

    # this is not working in meshing mode
    # assert on_case_loaded.loaded


@pytest.mark.fluent_version(">=23.1")
def test_iteration_ended_sync_event(static_mixer_case_session):
    solver = static_mixer_case_session
    solver.settings.solution.initialization.hybrid_initialize()
    count = 0

    def cb(session, event_info):
        assert event_info.index == session.scheme_eval.scheme_eval(
            "(get-current-iteration)"
        )
        nonlocal count
        count += 1

    cb_id = solver.events.register_callback(pyfluent.SolverEvent.ITERATION_ENDED, cb)
    solver.settings.solution.run_calculation.iterate(iter_count=10)
    assert count == 10
    solver.events.unregister_callback(cb_id)
    solver.settings.solution.run_calculation.iterate(iter_count=5)
    assert count == 10


@pytest.mark.fluent_version(">=23.1")
def test_iteration_ended_sync_event_multiple_connections(static_mixer_case_session):
    solver1 = static_mixer_case_session
    solver2 = pyfluent.connect_to_fluent(
        ip=solver1.connection_properties.ip,
        port=solver1.connection_properties.port,
        password=solver1.connection_properties.password,
    )
    solver1.settings.solution.initialization.hybrid_initialize()
    solver1_count = 0
    solver2_count = 0

    def solver1_cb(session, event_info):
        assert session == solver1
        if event_info.index % 2 == 0:
            nonlocal solver1_count
            solver1_count += 1

    def solver2_cb(session, event_info):
        assert session == solver2
        if event_info.index % 2 == 1:
            nonlocal solver2_count
            solver2_count += 1

    solver1.events.register_callback(pyfluent.SolverEvent.ITERATION_ENDED, solver1_cb)
    solver2.events.register_callback(pyfluent.SolverEvent.ITERATION_ENDED, solver2_cb)
    solver2.settings.solution.run_calculation.iterate(iter_count=5)
    assert solver1_count == 2
    assert solver2_count == 3


@pytest.mark.fluent_version(">=23.1")
def test_timestep_ended_sync_event(static_mixer_case_session):
    solver = static_mixer_case_session
    solver.settings.setup.general.solver.time = "unsteady-2nd-order"
    solver.settings.solution.initialization.hybrid_initialize()
    count = 0

    def cb(session, event_info):
        assert event_info.index == solver.rp_vars("time-step")
        nonlocal count
        count += 1

    cb_id = solver.events.register_callback(pyfluent.SolverEvent.TIMESTEP_ENDED, cb)
    solver.settings.solution.run_calculation.dual_time_iterate(
        time_step_count=10, max_iter_per_step=2
    )
    assert count == 10
    solver.events.unregister_callback(cb_id)
    solver.settings.solution.run_calculation.dual_time_iterate(
        time_step_count=5, max_iter_per_step=2
    )
    assert count == 10


@pytest.mark.fluent_version(">=23.1")
def test_sync_event_exception_in_callback(static_mixer_case_session, caplog):
    solver = static_mixer_case_session
    solver.settings.solution.initialization.hybrid_initialize()

    def cb(session, event_info):
        raise RuntimeError

    cb_id = solver.events.register_callback(pyfluent.SolverEvent.ITERATION_ENDED, cb)
    with caplog.at_level("ERROR", logger="pyfluent.networking"):
        solver.settings.solution.run_calculation.iterate(iter_count=10)
    assert (
        len(
            [
                record
                for record in caplog.records
                if "Error in callback" in record.message
            ]
        )
        == 10
    )
    caplog.clear()
    solver.events.unregister_callback(cb_id)
    with caplog.at_level("ERROR", logger="pyfluent.networking"):
        solver.settings.solution.run_calculation.iterate(iter_count=5)
    assert (
        len(
            [
                record
                for record in caplog.records
                if "Error in callback" in record.message
            ]
        )
        == 0
    )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
from typing import Iterable

import pytest

from ansys.fluent.core import FluentVersion, examples
from ansys.fluent.core.workflow import camel_to_snake_case


@pytest.mark.nightly
@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_new_watertight_workflow(new_meshing_session):
    # Import geometry
    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    watertight = new_meshing_session.watertight()
    watertight.import_geometry.file_name.set_state(import_file_name)
    assert watertight.import_geometry.length_unit() == "mm"
    watertight.import_geometry.length_unit.set_state("in")
    assert watertight.import_geometry.length_unit.get_state() == "in"
    watertight.import_geometry()

    # Add local sizing
    watertight.add_local_sizing.add_child_to_task()
    watertight.add_local_sizing()

    # Generate surface mesh
    watertight.create_surface_mesh.cfd_surface_mesh_controls.max_size.set_state(0.3)
    assert watertight.create_surface_mesh.cfd_surface_mesh_controls.max_size() == 0.3
    watertight.create_surface_mesh()

    # Describe geometry
    watertight.describe_geometry.update_child_tasks(setup_type_changed=False)
    watertight.describe_geometry.setup_type.set_state(
        "The geometry consists of only fluid regions with no voids"
    )
    watertight.describe_geometry.update_child_tasks(setup_type_changed=True)
    watertight.describe_geometry()

    # Update boundaries
    watertight.update_boundaries.boundary_label_list.set_state(["wall-inlet"])
    watertight.update_boundaries.boundary_label_type_list.set_state(["wall"])
    watertight.update_boundaries.old_boundary_label_list.set_state(["wall-inlet"])
    watertight.update_boundaries.old_boundary_label_type_list.set_state(
        ["velocity-inlet"]
    )
    watertight.update_boundaries()

    # Update regions
    watertight.update_regions()

    # Add boundary layers
    watertight.add_boundary_layer.add_child_to_task()
    watertight.add_boundary_layer.insert_compound_child_task()
    watertight.add_boundary_layer.arguments = {}
    watertight.add_boundary_layer_child_1.bl_control_name.set_state(
        "smooth-transition_1"
    )
    watertight.add_boundary_layer_child_1()

    # Generate volume mesh
    watertight.create_volume_mesh.volume_fill.set_state("poly-hexcore")
    watertight.create_volume_mesh.volume_fill_controls.hex_max_cell_length.set_state(
        0.3
    )
    watertight.create_volume_mesh()

    # Switch to solution mode
    solver = new_meshing_session.switch_to_solver()
    assert solver


@pytest.mark.nightly
@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_new_fault_tolerant_workflow(new_meshing_session):
    meshing = new_meshing_session

    # Import CAD and part management
    import_file_name = examples.download_file(
        "exhaust_system.fmd", "pyfluent/exhaust_system"
    )
    fault_tolerant = meshing.fault_tolerant()
    meshing.PartManagement.InputFileChanged(
        FilePath=import_file_name, IgnoreSolidNames=False, PartPerBody=False
    )
    meshing.PMFileManagement.FileManager.LoadFiles()
    meshing.PartManagement.Node["Meshing Model"].Copy(
        Paths=[
            "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/main,1",
            "/dirty_manifold-for-wrapper,"
            + "1/dirty_manifold-for-wrapper,1/flow-pipe,1",
            "/dirty_manifold-for-wrapper,"
            + "1/dirty_manifold-for-wrapper,1/outpipe3,1",
            "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/object2,1",
            "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/object1,1",
        ]
    )
    meshing.PartManagement.ObjectSetting["DefaultObjectSetting"].OneZonePer.set_state(
        "part"
    )
    fault_tolerant.import_cad_and_part_management.context.set_state(0)
    fault_tolerant.import_cad_and_part_management.create_object_per.set_state("Custom")
    fault_tolerant.import_cad_and_part_management.fmd_file_name.set_state(
        import_file_name
    )
    fault_tolerant.import_cad_and_part_management.file_loaded.set_state("yes")
    fault_tolerant.import_cad_and_part_management.object_setting.set_state(
        "DefaultObjectSetting"
    )
    fault_tolerant.import_cad_and_part_management.options.line.set_state(False)
    fault_tolerant.import_cad_and_part_management.options.solid.set_state(False)
    fault_tolerant.import_cad_and_part_management.options.surface.set_state(False)
    fault_tolerant.import_cad_and_part_management()

    # Describe geometry and flow
    fault_tolerant.describe_geometry_and_flow.add_enclosure.set_state("No")
    fault_tolerant.describe_geometry_and_flow.close_caps.set_state("Yes")
    fault_tolerant.describe_geometry_and_flow.flow_type.set_state(
        "Internal flow through the object"
    )
    fault_tolerant.describe_geometry_and_flow.update_child_tasks(
        setup_type_changed=False
    )

    fault_tolerant.describe_geometry_and_flow.add_enclosure.set_state("No")
    fault_tolerant.describe_geometry_and_flow.close_caps.set_state("Yes")
    fault_tolerant.describe_geometry_and_flow.describe_geometry_and_flow_options.advanced_options.set_state(
        True
    )
    fault_tolerant.describe_geometry_and_flow.describe_geometry_and_flow_options.extract_edge_features.set_state(
        "Yes"
    )
    fault_tolerant.describe_geometry_and_flow.flow_type.set_state(
        "Internal flow through the object"
    )
    fault_tolerant.describe_geometry_and_flow.update_child_tasks(
        setup_type_changed=False
    )
    fault_tolerant.describe_geometry_and_flow()

    # Enclose fluid regions (capping)
    fault_tolerant.enclose_fluid_regions_fault.create_patch_preferences.show_create_patch_preferences.set_state(
        False
    )
    fault_tolerant.enclose_fluid_regions_fault.patch_name.set_state("inlet-1")
    fault_tolerant.enclose_fluid_regions_fault.selection_type.set_state("zone")
    fault_tolerant.enclose_fluid_regions_fault.zone_selection_list.set_state(
        ["inlet.1"]
    )

    fault_tolerant.enclose_fluid_regions_fault.create_patch_preferences.show_create_patch_preferences.set_state(
        False
    )
    fault_tolerant.enclose_fluid_regions_fault.patch_name.set_state("inlet-1")
    fault_tolerant.enclose_fluid_regions_fault.selection_type.set_state("zone")
    fault_tolerant.enclose_fluid_regions_fault.zone_location.set_state(
        [
            "1",
            "351.68205",
            "-361.34322",
            "-301.88668",
            "396.96205",
            "-332.84759",
            "-266.69751",
            "inlet.1",
        ]
    )
    fault_tolerant.enclose_fluid_regions_fault.zone_selection_list.set_state(
        ["inlet.1"]
    )
    fault_tolerant.enclose_fluid_regions_fault.add_child_to_task()
    fault_tolerant.enclose_fluid_regions_fault.insert_compound_child_task()
    fault_tolerant.enclose_fluid_regions_fault.arguments.set_state({})
    fault_tolerant.enclose_fluid_regions_fault_child_1()

    fault_tolerant.enclose_fluid_regions_fault.patch_name.set_state("inlet-2")
    fault_tolerant.enclose_fluid_regions_fault.selection_type.set_state("zone")
    fault_tolerant.enclose_fluid_regions_fault.zone_selection_list.set_state(
        ["inlet.2"]
    )

    fault_tolerant.enclose_fluid_regions_fault.patch_name.set_state("inlet-2")
    fault_tolerant.enclose_fluid_regions_fault.selection_type.set_state("zone")
    fault_tolerant.enclose_fluid_regions_fault.zone_location.set_state(
        [
            "1",
            "441.68205",
            "-361.34322",
            "-301.88668",
            "486.96205",
            "-332.84759",
            "-266.69751",
            "inlet.2",
        ]
    )
    fault_tolerant.enclose_fluid_regions_fault.zone_selection_list.set_state(
        ["inlet.2"]
    )
    fault_tolerant.enclose_fluid_regions_fault.add_child_to_task()
    fault_tolerant.enclose_fluid_regions_fault.insert_compound_child_task()
    fault_tolerant.enclose_fluid_regions_fault.arguments.set_state({})
    fault_tolerant.enclose_fluid_regions_fault_child_2()

    fault_tolerant.enclose_fluid_regions_fault.patch_name.set_state("inlet-3")
    fault_tolerant.enclose_fluid_regions_fault.selection_type.set_state("zone")
    fault_tolerant.enclose_fluid_regions_fault.zone_selection_list.set_state(["inlet"])

    fault_tolerant.enclose_fluid_regions_fault.patch_name.set_state("inlet-3")
    fault_tolerant.enclose_fluid_regions_fault.selection_type.set_state("zone")
    fault_tolerant.enclose_fluid_regions_fault.zone_location.set_state(
        [
            "1",
            "261.68205",
            "-361.34322",
            "-301.88668",
            "306.96205",
            "-332.84759",
            "-266.69751",
            "inlet",
        ]
    )
    fault_tolerant.enclose_fluid_regions_fault.zone_selection_list.set_state(["inlet"])
    fault_tolerant.enclose_fluid_regions_fault.add_child_to_task()
    fault_tolerant.enclose_fluid_regions_fault.insert_compound_child_task()
    fault_tolerant.enclose_fluid_regions_fault.arguments.set_state({})
    fault_tolerant.enclose_fluid_regions_fault_child_3()

    fault_tolerant.enclose_fluid_regions_fault.patch_name.set_state("outlet-1")
    fault_tolerant.enclose_fluid_regions_fault.selection_type.set_state("zone")
    fault_tolerant.enclose_fluid_regions_fault.zone_selection_list.set_state(["outlet"])
    fault_tolerant.enclose_fluid_regions_fault.zone_type.set_state("pressure-outlet")

    fault_tolerant.enclose_fluid_regions_fault.patch_name.set_state("outlet-1")
    fault_tolerant.enclose_fluid_regions_fault.selection_type.set_state("zone")
    fault_tolerant.enclose_fluid_regions_fault.zone_location.set_state(
        [
            "1",
            "352.22702",
            "-197.8957",
            "84.102381",
            "394.41707",
            "-155.70565",
            "84.102381",
            "outlet",
        ]
    )
    fault_tolerant.enclose_fluid_regions_fault.zone_selection_list.set_state(["outlet"])
    fault_tolerant.enclose_fluid_regions_fault.zone_type.set_state("pressure-outlet")
    fault_tolerant.enclose_fluid_regions_fault.add_child_to_task()
    fault_tolerant.enclose_fluid_regions_fault.insert_compound_child_task()
    fault_tolerant.enclose_fluid_regions_fault.arguments.set_state({})
    fault_tolerant.enclose_fluid_regions_fault_child_4()

    # Extract edge features
    fault_tolerant.extract_edge_features.extract_method_type.set_state(
        "Intersection Loops"
    )
    fault_tolerant.extract_edge_features.object_selection_list.set_state(
        ["flow_pipe", "main"]
    )
    fault_tolerant.extract_edge_features.add_child_to_task()
    fault_tolerant.extract_edge_features.insert_compound_child_task()

    fault_tolerant.extract_edge_features.extract_edges_name.set_state("edge-group-1")
    fault_tolerant.extract_edge_features.extract_method_type.set_state(
        "Intersection Loops"
    )
    fault_tolerant.extract_edge_features.object_selection_list.set_state(
        ["flow_pipe", "main"]
    )

    fault_tolerant.extract_edge_features.arguments.set_state({})
    fault_tolerant.extract_edge_features_child_1()

    # Identify regions
    fault_tolerant.identify_regions.selection_type.set_state("zone")
    fault_tolerant.identify_regions.x.set_state(377.322045740589)
    fault_tolerant.identify_regions.y.set_state(-176.800676988458)
    fault_tolerant.identify_regions.z.set_state(-37.0764628583475)
    fault_tolerant.identify_regions.zone_selection_list.set_state(["main.1"])

    fault_tolerant.identify_regions.selection_type.set_state("zone")
    fault_tolerant.identify_regions.x.set_state(377.322045740589)
    fault_tolerant.identify_regions.y.set_state(-176.800676988458)
    fault_tolerant.identify_regions.z.set_state(-37.0764628583475)
    fault_tolerant.identify_regions.zone_location.set_state(
        [
            "1",
            "213.32205",
            "-225.28068",
            "-158.25531",
            "541.32205",
            "-128.32068",
            "84.102381",
            "main.1",
        ]
    )
    fault_tolerant.identify_regions.zone_selection_list.set_state(["main.1"])
    fault_tolerant.identify_regions.add_child_to_task()
    fault_tolerant.identify_regions.insert_compound_child_task()
    fault_tolerant.identify_regions.x.set_state(377.322045740589)
    fault_tolerant.identify_regions.y.set_state(-176.800676988458)
    fault_tolerant.identify_regions.z.set_state(-37.0764628583475)

    fault_tolerant.identify_regions_child_1.material_points_name.set_state(
        "fluid-region-1"
    )
    fault_tolerant.identify_regions_child_1.selection_type.set_state("zone")
    fault_tolerant.identify_regions.zone_location.set_state(
        [
            "1",
            "213.32205",
            "-225.28068",
            "-158.25531",
            "541.32205",
            "-128.32068",
            "84.102381",
            "main.1",
        ]
    )
    fault_tolerant.identify_regions.zone_selection_list.set_state(["main.1"])
    fault_tolerant.identify_regions.arguments.set_state({})
    fault_tolerant.identify_regions_child_1()

    fault_tolerant.identify_regions.material_points_name.set_state("void-region-1")
    fault_tolerant.identify_regions.new_region_type.set_state("void")
    fault_tolerant.identify_regions.object_selection_list.set_state(
        ["inlet-1", "inlet-2", "inlet-3", "main"]
    )
    fault_tolerant.identify_regions.x.set_state(374.722045740589)
    fault_tolerant.identify_regions.y.set_state(-278.9775145640143)
    fault_tolerant.identify_regions.z.set_state(-161.1700719416913)
    fault_tolerant.identify_regions.add_child_to_task()
    fault_tolerant.identify_regions.insert_compound_child_task()
    fault_tolerant.identify_regions.arguments.set_state({})
    fault_tolerant.identify_regions_child_2()

    # Define leakage threshold
    fault_tolerant.define_leakage_threshold.add_child.set_state("yes")
    fault_tolerant.define_leakage_threshold.flip_direction.set_state(True)
    fault_tolerant.define_leakage_threshold.plane_direction.set_state("X")
    fault_tolerant.define_leakage_threshold.region_selection_single.set_state(
        "void-region-1"
    )
    fault_tolerant.define_leakage_threshold.add_child_to_task()
    fault_tolerant.define_leakage_threshold.insert_compound_child_task()

    fault_tolerant.define_leakage_threshold.add_child.set_state("yes")

    fault_tolerant.define_leakage_threshold_child_1.arguments.set_state(
        {
            "add_child": "yes",
            "flip_direction": True,
            "leakage_name": "leakage-1",
            "plane_direction": "X",
            "region_selection_single": "void-region-1",
        }
    )

    fault_tolerant.define_leakage_threshold_child_1()

    # Update regions settings
    fault_tolerant.update_region_settings.all_region_filter_categories.set_state(
        ["2"] * 5 + ["1"] * 2
    )
    fault_tolerant.update_region_settings.all_region_leakage_size_list.set_state(
        ["none"] * 6 + ["6.4"]
    )
    fault_tolerant.update_region_settings.all_region_linked_construction_surface_list.set_state(
        ["n/a"] * 6 + ["no"]
    )
    fault_tolerant.update_region_settings.all_region_mesh_method_list.set_state(
        ["none"] * 6 + ["wrap"]
    )
    fault_tolerant.update_region_settings.all_region_name_list.set_state(
        [
            "main",
            "flow_pipe",
            "outpipe3",
            "object2",
            "object1",
            "void-region-1",
            "fluid-region-1",
        ]
    )
    fault_tolerant.update_region_settings.all_region_overset_componen_list.set_state(
        ["no"] * 7
    )
    fault_tolerant.update_region_settings.all_region_source_list.set_state(
        ["object"] * 5 + ["mpt"] * 2
    )
    fault_tolerant.update_region_settings.all_region_type_list.set_state(
        ["void"] * 6 + ["fluid"]
    )
    fault_tolerant.update_region_settings.all_region_volume_fill_list.set_state(
        ["none"] * 6 + ["tet"]
    )
    fault_tolerant.update_region_settings.filter_category.set_state(
        "Identified Regions"
    )
    fault_tolerant.update_region_settings.old_region_leakage_size_list.set_state([""])
    fault_tolerant.update_region_settings.old_region_mesh_method_list.set_state(
        ["wrap"]
    )
    fault_tolerant.update_region_settings.old_region_name_list.set_state(
        ["fluid-region-1"]
    )
    fault_tolerant.update_region_settings.old_region_overset_componen_list.set_state(
        ["no"]
    )
    fault_tolerant.update_region_settings.old_region_type_list.set_state(["fluid"])
    fault_tolerant.update_region_settings.old_region_volume_fill_list.set_state(
        ["hexcore"]
    )
    fault_tolerant.update_region_settings.region_leakage_size_list.set_state([""])
    fault_tolerant.update_region_settings.region_mesh_method_list.set_state(["wrap"])
    fault_tolerant.update_region_settings.region_name_list.set_state(["fluid-region-1"])
    fault_tolerant.update_region_settings.region_overset_componen_list.set_state(["no"])
    fault_tolerant.update_region_settings.region_type_list.set_state(["fluid"])
    fault_tolerant.update_region_settings.region_volume_fill_list.set_state(["tet"])
    fault_tolerant.update_region_settings()

    # Choose mesh control options
    fault_tolerant.choose_mesh_control_options()

    # Generate surface mesh
    if meshing.get_fluent_version() < FluentVersion.v251:
        fault_tolerant.generate_the_surface_mesh()
    else:
        fault_tolerant.generate_surface_mesh()

    # Update boundaries
    fault_tolerant.update_boundaries_ftm()

    # Add boundary layers
    fault_tolerant.add_boundary_layer_ftm.add_child_to_task()
    fault_tolerant.add_boundary_layer_ftm.insert_compound_child_task()
    fault_tolerant.add_boundary_layer_ftm.arguments.set_state({})
    fault_tolerant.add_boundary_layer_ftm_child_1.bl_control_name.set_state(
        "aspect-ratio_1"
    )
    fault_tolerant.add_boundary_layer_ftm_child_1()

    # Generate volume mesh
    if meshing.get_fluent_version() < FluentVersion.v251:
        generate_volume_mesh = fault_tolerant.generate_the_volume_mesh
    else:
        generate_volume_mesh = fault_tolerant.create_volume_mesh
    generate_volume_mesh.all_region_name_list.set_state(
        [
            "main",
            "flow_pipe",
            "outpipe3",
            "object2",
            "object1",
            "void-region-1",
            "fluid-region-1",
        ]
    )
    generate_volume_mesh.all_region_size_list.set_state(["11.33375"] * 7)
    generate_volume_mesh.all_region_volume_fill_list.set_state(["none"] * 6 + ["tet"])
    generate_volume_mesh.enable_parallel.set_state(True)
    generate_volume_mesh()

    # Generate volume mesh
    solver = meshing.switch_to_solver()
    assert solver


@pytest.mark.nightly
@pytest.mark.codegen_required
@pytest.mark.fluent_version("==24.2")
def test_new_2d_meshing_workflow(new_meshing_session):
    # Import geometry
    import_file_name = examples.download_file("NACA0012.fmd", "pyfluent/airfoils")
    two_dim_mesh = new_meshing_session.two_dimensional_meshing()

    two_dim_mesh.load_cad_geometry_2d.file_name = import_file_name
    two_dim_mesh.load_cad_geometry_2d.length_unit = "mm"
    two_dim_mesh.load_cad_geometry_2d.refaceting.refacet = False
    two_dim_mesh.load_cad_geometry_2d()

    # Set regions and boundaries
    two_dim_mesh.update_regions_2d()
    two_dim_mesh.update_boundaries_2d.selection_type = "zone"
    two_dim_mesh.update_boundaries_2d()

    # Define global sizing
    two_dim_mesh.define_global_sizing_2d.curvature_normal_angle = 20
    two_dim_mesh.define_global_sizing_2d.max_size = 2000.0
    two_dim_mesh.define_global_sizing_2d.min_size = 5.0
    two_dim_mesh.define_global_sizing_2d.size_functions = "Curvature"
    two_dim_mesh.define_global_sizing_2d()

    # Add local sizing
    two_dim_mesh.add_local_sizing_2d.add_child = "yes"
    two_dim_mesh.add_local_sizing_2d.boi_control_name = "boi_1"
    two_dim_mesh.add_local_sizing_2d.boi_execution = "Body Of Influence"
    two_dim_mesh.add_local_sizing_2d.boi_face_label_list = ["boi"]
    two_dim_mesh.add_local_sizing_2d.boi_size = 50.0
    two_dim_mesh.add_local_sizing_2d.boi_zoneor_label = "label"
    two_dim_mesh.add_local_sizing_2d.draw_size_control = True
    two_dim_mesh.add_local_sizing_2d.add_child_and_update(defer_update=False)

    two_dim_mesh.add_local_sizing_2d.add_child = "yes"
    two_dim_mesh.add_local_sizing_2d.boi_control_name = "edgesize_1"
    two_dim_mesh.add_local_sizing_2d.boi_execution = "Edge Size"
    two_dim_mesh.add_local_sizing_2d.boi_size = 5.0
    two_dim_mesh.add_local_sizing_2d.boi_zoneor_label = "label"
    two_dim_mesh.add_local_sizing_2d.draw_size_control = True
    two_dim_mesh.add_local_sizing_2d.edge_label_list = ["airfoil-te"]
    two_dim_mesh.add_local_sizing_2d.add_child_and_update(defer_update=False)

    two_dim_mesh.add_local_sizing_2d.add_child = "yes"
    two_dim_mesh.add_local_sizing_2d.boi_control_name = "curvature_1"
    two_dim_mesh.add_local_sizing_2d.boi_curvature_normal_angle = 10
    two_dim_mesh.add_local_sizing_2d.boi_execution = "Curvature"
    two_dim_mesh.add_local_sizing_2d.boi_max_size = 2
    two_dim_mesh.add_local_sizing_2d.boi_min_size = 1.5
    two_dim_mesh.add_local_sizing_2d.boi_scope_to = "edges"
    two_dim_mesh.add_local_sizing_2d.boi_zoneor_label = "label"
    two_dim_mesh.add_local_sizing_2d.draw_size_control = True
    two_dim_mesh.add_local_sizing_2d.edge_label_list = ["airfoil"]
    two_dim_mesh.add_local_sizing_2d.add_child_and_update(defer_update=False)

    # Add boundary layer
    two_dim_mesh.add_2d_boundary_layers.add_child = "yes"
    two_dim_mesh.add_2d_boundary_layers.bl_control_name = "aspect-ratio_1"
    two_dim_mesh.add_2d_boundary_layers.number_of_layers = 4
    two_dim_mesh.add_2d_boundary_layers.offset_method_type = "aspect-ratio"
    two_dim_mesh.add_2d_boundary_layers.add_child_and_update(defer_update=False)

    two_dim_mesh.generate_initial_surface_mesh.surface_2d_preferences.merge_edge_zones_based_on_labels = (
        "no"
    )
    two_dim_mesh.generate_initial_surface_mesh.surface_2d_preferences.merge_face_zones_based_on_labels = (
        "no"
    )
    two_dim_mesh.generate_initial_surface_mesh.surface_2d_preferences.show_advanced_options = (
        True
    )
    two_dim_mesh.generate_initial_surface_mesh()

    two_dim_mesh.add_2d_boundary_layers_child_1.revert()
    two_dim_mesh.add_2d_boundary_layers_child_1.add_child = "yes"
    two_dim_mesh.add_2d_boundary_layers_child_1.bl_control_name = "uniform_1"
    two_dim_mesh.add_2d_boundary_layers_child_1.first_layer_height = 2
    two_dim_mesh.add_2d_boundary_layers_child_1.number_of_layers = 4
    two_dim_mesh.add_2d_boundary_layers_child_1.offset_method_type = "uniform"
    two_dim_mesh.add_2d_boundary_layers_child_1()

    two_dim_mesh.generate_initial_surface_mesh.surface_2d_preferences.merge_edge_zones_based_on_labels = (
        "no"
    )
    two_dim_mesh.generate_initial_surface_mesh.surface_2d_preferences.merge_face_zones_based_on_labels = (
        "no"
    )
    two_dim_mesh.generate_initial_surface_mesh.surface_2d_preferences.show_advanced_options = (
        True
    )
    two_dim_mesh.generate_initial_surface_mesh()

    two_dim_mesh._task("uniform_1").revert()
    two_dim_mesh._task("uniform_1").add_child = "yes"
    two_dim_mesh._task("uniform_1").bl_control_name = "smooth-transition_1"
    two_dim_mesh._task("uniform_1").first_layer_height = 2
    two_dim_mesh._task("uniform_1").number_of_layers = 7
    two_dim_mesh._task("uniform_1").offset_method_type = "smooth-transition"
    two_dim_mesh._task("uniform_1")()

    two_dim_mesh.generate_initial_surface_mesh.surface_2d_preferences.merge_edge_zones_based_on_labels = (
        "no"
    )
    two_dim_mesh.generate_initial_surface_mesh.surface_2d_preferences.merge_face_zones_based_on_labels = (
        "no"
    )
    two_dim_mesh.generate_initial_surface_mesh.surface_2d_preferences.show_advanced_options = (
        True
    )
    two_dim_mesh.generate_initial_surface_mesh()

    # Switch to solution mode
    solver = new_meshing_session.switch_to_solver()
    assert solver


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_updating_state_in_new_meshing_workflow(new_meshing_session):
    # Import geometry
    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    watertight = new_meshing_session.watertight()
    assert watertight.import_geometry.length_unit() == "mm"
    assert watertight.import_geometry.cad_import_options.feature_angle() == 40.0
    assert (
        watertight.import_geometry.cad_import_options.one_zone_per.allowed_values()
        == ["body", "face", "object"]
    )
    assert watertight.import_geometry.cad_import_options.one_zone_per() == "body"
    watertight.import_geometry.arguments = {
        "file_name": import_file_name,
        "length_unit": "in",
        "cad_import_options": {"feature_angle": 35, "one_zone_per": "object"},
    }
    assert watertight.import_geometry.cad_import_options.feature_angle() == 35.0
    assert (
        watertight.import_geometry.cad_import_options.one_zone_per.get_state()
        == "object"
    )
    assert watertight.import_geometry.length_unit.get_state() == "in"
    watertight.import_geometry.cad_import_options.feature_angle = 25.0
    assert watertight.import_geometry.cad_import_options.feature_angle() == 25.0
    watertight.import_geometry.cad_import_options.one_zone_per = "face"
    assert watertight.import_geometry.cad_import_options.one_zone_per() == "face"
    watertight.import_geometry()


def _assert_snake_case_attrs(attrs: Iterable):
    for attr in attrs:
        assert str(attr).islower()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_snake_case_attrs_in_new_meshing_workflow(new_meshing_session):
    # Import geometry
    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    watertight = new_meshing_session.watertight()
    dir_watertight = dir(watertight)
    dir_watertight.remove("_FirstTask")
    _assert_snake_case_attrs(dir_watertight)
    dir_watertight_import_geometry = dir(watertight.import_geometry)
    dir_watertight_import_geometry.remove("_NextTask")
    _assert_snake_case_attrs(dir_watertight_import_geometry)
    _assert_snake_case_attrs(watertight.import_geometry.arguments())
    _assert_snake_case_attrs(watertight.import_geometry.cad_import_options())
    _assert_snake_case_attrs(dir(watertight.import_geometry.cad_import_options))
    watertight.import_geometry.file_name.set_state(import_file_name)
    watertight.import_geometry.length_unit = "in"
    watertight.import_geometry()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_workflow_and_data_model_methods_new_meshing_workflow(new_meshing_session):
    # Import geometry
    meshing = new_meshing_session
    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    watertight = meshing.watertight()

    # Checks if any of the unwanted attrs are present in dir call
    assert (set(dir(watertight)) - watertight._unwanted_attrs) == set(dir(watertight))

    for attr in watertight._unwanted_attrs:
        with pytest.raises(AttributeError):
            getattr(watertight, attr)

    watertight.import_geometry.rename(new_name="import_geom_wtm")
    assert "import_geometry" not in watertight.task_names()
    assert "import_geom_wtm" in watertight.task_names()
    assert len(watertight.tasks()) == 11
    watertight.import_geom_wtm.file_name = import_file_name
    watertight.import_geom_wtm.length_unit = "in"
    watertight.import_geom_wtm()
    _next_possible_tasks = [
        "<Insertable 'import_boi_geometry' task>",
        "<Insertable 'set_up_rotational_periodic_boundaries' task>",
        "<Insertable 'create_local_refinement_regions' task>",
        "<Insertable 'custom_journal_task' task>",
    ]
    assert sorted(
        [repr(x) for x in watertight.import_geom_wtm.insertable_tasks()]
    ) == sorted(_next_possible_tasks)
    watertight.import_geom_wtm.insertable_tasks.import_boi_geometry.insert()
    assert sorted(
        [repr(x) for x in watertight.import_geom_wtm.insertable_tasks()]
    ) == sorted(_next_possible_tasks)
    watertight.import_geom_wtm.insertable_tasks.set_up_rotational_periodic_boundaries.insert()
    assert len(watertight.tasks()) == 13


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_watertight_workflow(mixing_elbow_geometry_filename, new_meshing_session):
    watertight = new_meshing_session.watertight()
    watertight.import_geometry.file_name = mixing_elbow_geometry_filename
    watertight.import_geometry()
    add_local_sizing = watertight.add_local_sizing
    assert not add_local_sizing.tasks()
    add_local_sizing._add_child(state={"boi_face_label_list": ["cold-inlet"]})
    assert not add_local_sizing.tasks()
    added_sizing = add_local_sizing.add_child_and_update(
        state={"boi_face_label_list": ["elbow-fluid"]}
    )
    assert len(add_local_sizing.tasks()) == 1
    assert added_sizing
    assert added_sizing.boi_face_label_list() == ["elbow-fluid"]


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_watertight_workflow_children(
    mixing_elbow_geometry_filename, new_meshing_session
):
    watertight = new_meshing_session.watertight()
    watertight.import_geometry.file_name = mixing_elbow_geometry_filename
    watertight.import_geometry()
    add_local_sizing = watertight.add_local_sizing
    assert not add_local_sizing.tasks()
    add_local_sizing._add_child(state={"boi_face_label_list": ["cold-inlet"]})
    assert not add_local_sizing.tasks()
    added_sizing = add_local_sizing.add_child_and_update(
        state={"boi_face_label_list": ["elbow-fluid"]}
    )
    assert len(add_local_sizing.tasks()) == 1
    assert added_sizing
    assert added_sizing.boi_face_label_list() == ["elbow-fluid"]
    assert added_sizing.name() == "facesize_1"
    assert len(added_sizing.arguments())
    added_sizing_by_name = add_local_sizing.compound_child("facesize_1")
    added_sizing_by_pos = add_local_sizing.last_child()
    assert added_sizing.arguments() == added_sizing_by_name.arguments()
    assert added_sizing.arguments() == added_sizing_by_pos.arguments()
    assert added_sizing.python_name() == "add_local_sizing_child_1"
    describe_geometry = watertight.describe_geometry
    describe_geometry_children = describe_geometry.tasks()
    assert len(describe_geometry_children) == 2
    describe_geometry_child_task_python_names = describe_geometry.task_names()
    assert describe_geometry_child_task_python_names == [
        "enclose_fluid_regions",
        "create_regions",
    ]


@pytest.mark.fluent_version(">=24.1")
@pytest.mark.codegen_required
def test_watertight_workflow_dynamic_interface(
    mixing_elbow_geometry_filename, new_meshing_session
):
    watertight = new_meshing_session.watertight()
    watertight.import_geometry.file_name = mixing_elbow_geometry_filename
    watertight.import_geometry()
    create_volume_mesh = watertight.create_volume_mesh
    assert create_volume_mesh is not None
    watertight.delete_tasks(list_of_tasks=["create_volume_mesh"])
    assert "create_volume_mesh" not in watertight.task_names()

    assert sorted(
        [repr(x) for x in watertight.add_boundary_layer.insertable_tasks()]
    ) == sorted(
        [
            "<Insertable 'add_boundary_type' task>",
            "<Insertable 'update_boundaries' task>",
            "<Insertable 'set_up_rotational_periodic_boundaries' task>",
            "<Insertable 'modify_mesh_refinement' task>",
            "<Insertable 'improve_surface_mesh' task>",
            "<Insertable 'create_volume_mesh' task>",
            "<Insertable 'manage_zones_ftm' task>",
            "<Insertable 'update_regions' task>",
            "<Insertable 'custom_journal_task' task>",
        ]
    )
    watertight.add_boundary_layer.insertable_tasks.create_volume_mesh.insert()
    assert "create_volume_mesh" in watertight.task_names()
    create_volume_mesh = watertight.create_volume_mesh
    assert create_volume_mesh is not None

    assert (
        watertight.describe_geometry.create_regions.arguments()[
            "number_of_flow_volumes"
        ]
        == 1
    )
    watertight.delete_tasks(list_of_tasks=["create_regions"])
    assert "create_regions" not in watertight.task_names()
    assert watertight.describe_geometry.enclose_fluid_regions
    watertight.describe_geometry.enclose_fluid_regions.delete()
    assert "enclose_fluid_regions" not in watertight.task_names()
    watertight.create_volume_mesh.delete()
    assert "create_volume_mesh" not in watertight.task_names()


@pytest.mark.fluent_version("==23.2")
@pytest.mark.codegen_required
def test_fault_tolerant_workflow(exhaust_system_geometry_filename, new_meshing_session):
    fault_tolerant = new_meshing_session.fault_tolerant()
    part_management = fault_tolerant.part_management
    file_name = exhaust_system_geometry_filename
    part_management.LoadFmdFile(FilePath=file_name)
    part_management.MoveCADComponentsToNewObject(
        Paths=[r"/Bottom,1", r"/Left,1", r"/Others,1", r"/Right,1", r"/Top,1"]
    )
    part_management.Node["Object"].Rename(NewName=r"Engine")
    import_cad = fault_tolerant.import_cad_and_part_management
    import_cad.Arguments.setState(
        {
            r"CreateObjectPer": r"Custom",
            r"FMDFileName": file_name,
            r"FileLoaded": r"yes",
            r"ObjectSetting": r"DefaultObjectSetting",
        }
    )
    import_cad()


@pytest.mark.fluent_version(">=23.2")
@pytest.mark.codegen_required
def test_extended_wrapper(new_meshing_session, mixing_elbow_geometry_filename):
    watertight = new_meshing_session.watertight()
    import_geometry = watertight.import_geometry
    assert import_geometry.Arguments() == {}
    import_geometry.Arguments = dict(FileName=mixing_elbow_geometry_filename)
    assert 8 < len(import_geometry.arguments.get_state()) < 15
    assert len(import_geometry.arguments.get_state(explicit_only=True)) == 1
    import_geometry.arguments.set_state(dict(file_name=None))
    time.sleep(5)
    assert import_geometry.arguments.get_state(explicit_only=True) == dict(
        file_name=None
    )
    assert import_geometry.arguments.get_state()["file_name"] is None
    import_geometry.arguments.set_state(dict(file_name=mixing_elbow_geometry_filename))
    time.sleep(5)
    assert import_geometry.arguments.get_state(explicit_only=True) == dict(
        file_name=mixing_elbow_geometry_filename
    )
    assert import_geometry.file_name() == mixing_elbow_geometry_filename
    import_geometry.file_name.set_state("bob")
    time.sleep(5)
    assert import_geometry.file_name() == "bob"
    import_geometry.file_name.set_state(mixing_elbow_geometry_filename)
    import_geometry()
    add_local_sizing = watertight.add_local_sizing
    assert not add_local_sizing.tasks()
    # new_meshing_session.workflow.TaskObject["Add Local Sizing"]._add_child(state={"BOIFaceLabelList": ["elbow-fluid"]})
    add_local_sizing._add_child(state={"boi_face_label_list": ["cold-inlet"]})
    assert not add_local_sizing.tasks()

    added_sizing = add_local_sizing.add_child_and_update(
        state={"boi_face_label_list": ["elbow-fluid"]}
    )
    assert len(add_local_sizing.tasks()) == 1
    assert added_sizing
    assert added_sizing.boi_face_label_list() == ["elbow-fluid"]
    # restart
    watertight = new_meshing_session.watertight()
    assert import_geometry.state() == "Out-of-date"
    import_geometry(FileName=mixing_elbow_geometry_filename, AppendMesh=False)
    assert import_geometry.state() == "Up-to-date"
    import_geometry_state = import_geometry.arguments()
    assert len(import_geometry_state) > 2


@pytest.mark.fluent_version(">=23.1")
@pytest.mark.skip
def test_meshing_workflow_structure(new_meshing_session):
    """
    o Workflow
    |
    |--o Import Geometry
    |
    |--o Add Local Sizing
    |
    |--o Generate the Surface Mesh
    |
    |--o Describe Geometry
    |  |
    |  |--o Enclose Fluid Regions (Capping)
    |  |
    |  |--o Create Regions
    |
    |--o Update Regions
    |
    |--o Add Boundary Layers
    |
    |--o Generate the Volume Mesh
    """
    w = new_meshing_session.workflow
    w.InitializeWorkflow(WorkflowType="Watertight Geometry")

    task_names = (
        "Import Geometry",
        "Add Local Sizing",
        "Generate the Surface Mesh",
        "Describe Geometry",
        "Enclose Fluid Regions (Capping)",
        "Create Regions",
        "Update Regions",
        "Add Boundary Layers",
        "Generate the Volume Mesh",
    )

    (
        import_geom,
        add_sizing,
        gen_surf_mesh,
        describe_geometry,
        cap,
        create_regions,
        update_regions,
        add_boundary_layers,
        gen_vol_mesh,
    ) = all_tasks = [w._task(name) for name in task_names]

    def upstream_names(task):
        return {upstream.name() for upstream in task.get_direct_upstream_tasks()}

    def downstream_names(task):
        return {downstream.name() for downstream in task.get_direct_downstream_tasks()}

    assert upstream_names(import_geom) == set()
    assert downstream_names(import_geom) == {
        "Generate the Surface Mesh",
        "Add Local Sizing",
    }

    assert upstream_names(add_sizing) == {"Import Geometry"}
    assert downstream_names(add_sizing) == {"Generate the Surface Mesh"}

    assert upstream_names(gen_surf_mesh) == {"Import Geometry", "Add Local Sizing"}
    assert downstream_names(gen_surf_mesh) == {
        "Describe Geometry",
        "Add Boundary Layers",
        "Generate the Volume Mesh",
    }

    assert upstream_names(describe_geometry) == {
        "Generate the Surface Mesh",
        "Add Boundary Layers",
    }
    assert downstream_names(describe_geometry) == {
        "Update Regions",
        "Add Boundary Layers",
        "Generate the Volume Mesh",
    }

    assert upstream_names(cap) == {
        "Describe Geometry",
        "Add Boundary Layers",
        "Generate the Surface Mesh",
    }
    assert downstream_names(cap) == {
        "Describe Geometry",
        "Add Boundary Layers",
        "Generate the Volume Mesh",
    }

    assert upstream_names(create_regions) == {
        "Describe Geometry",
        "Add Boundary Layers",
        "Generate the Surface Mesh",
    }
    assert downstream_names(create_regions) == {
        "Describe Geometry",
        "Add Boundary Layers",
        "Generate the Volume Mesh",
        "Update Regions",
    }

    assert upstream_names(update_regions) == {"Describe Geometry"}
    assert downstream_names(update_regions) == {"Generate the Volume Mesh"}

    assert upstream_names(add_boundary_layers) == {
        "Describe Geometry",
        "Generate the Surface Mesh",
    }
    assert downstream_names(add_boundary_layers) == {
        "Describe Geometry",
        "Generate the Volume Mesh",
    }

    assert upstream_names(gen_vol_mesh) == {
        "Update Regions",
        "Describe Geometry",
        "Add Boundary Layers",
        "Generate the Surface Mesh",
    }
    assert downstream_names(gen_vol_mesh) == set()

    for task in all_tasks:
        assert {sub_task.name() for sub_task in task.tasks()} == (
            {
                "Enclose Fluid Regions (Capping)",
                "Create Regions",
            }
            if task is describe_geometry
            else set()
        )

    for task in all_tasks:
        assert {sub_task.name() for sub_task in task.inactive_tasks()} == (
            {
                "Apply Share Topology",
                "Update Boundaries",
            }
            if task is describe_geometry
            else set()
        )

    task_ids = [task.get_id() for task in all_tasks]
    # uniqueness test
    assert len(set(task_ids)) == len(task_ids)
    # ordering test
    idxs = [int(id[len("TaskObject") :]) for id in task_ids]
    assert sorted(idxs) == idxs
    """Given the workflow::

            Workflow
            ├── Import Geometry
            ├── Add Local Sizing
            ├── Generate the Surface Mesh ── Insert Next Task
                                            ├── Add Boundary Type
                                            ├── Update Boundaries
                                            ├── ...
    """
    assert set(gen_surf_mesh.GetNextPossibleTasks()) == {
        "AddBoundaryType",
        "UpdateBoundaries",
        "SetUpPeriodicBoundaries",
        "LinearMeshPattern",
        "ManageZones",
        "ModifyMeshRefinement",
        "ImproveSurfaceMesh",
        "RunCustomJournal",
    }

    children = w.tasks()
    expected_task_order = (
        "Import Geometry",
        "Add Local Sizing",
        "Generate the Surface Mesh",
        "Describe Geometry",
        "Update Regions",
        "Add Boundary Layers",
        "Generate the Volume Mesh",
    )

    actual_task_order = tuple(child.name() for child in children)

    assert actual_task_order == expected_task_order

    assert [child.name() for child in children[3].tasks()] == [
        "Enclose Fluid Regions (Capping)",
        "Create Regions",
    ]

    gen_surf_mesh.InsertNextTask(CommandName="AddBoundaryType")

    children = w.tasks()
    expected_task_order = (
        "Import Geometry",
        "Add Local Sizing",
        "Generate the Surface Mesh",
        "Add Boundary Type",
        "Describe Geometry",
        "Update Regions",
        "Add Boundary Layers",
        "Generate the Volume Mesh",
    )

    actual_task_order = tuple(child.name() for child in children)

    assert actual_task_order == expected_task_order

    assert [child.name() for child in children[4].tasks()] == [
        "Enclose Fluid Regions (Capping)",
        "Create Regions",
    ]


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_new_workflow_structure(new_meshing_session):
    meshing = new_meshing_session
    watertight = meshing.watertight()
    assert watertight.import_geometry.arguments()
    with pytest.raises(AttributeError):
        watertight.TaskObject["Import Geometry"]


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_attrs_in_watertight_meshing_workflow(new_meshing_session):
    # Import geometry
    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    watertight = new_meshing_session.watertight()
    unwanted_attrs = {"fault_tolerant", "part_management", "pm_file_management"}
    assert set(dir(watertight)) - unwanted_attrs == set(dir(watertight))

    for attr in unwanted_attrs:
        with pytest.raises(AttributeError):
            getattr(watertight, attr)

    watertight.import_geometry.file_name.set_state(import_file_name)
    watertight.import_geometry.length_unit = "in"
    watertight.import_geometry()

    assert watertight.import_geometry.file_name()
    # Reinitialize the workflow:
    watertight = new_meshing_session.watertight()
    assert not watertight.import_geometry.file_name()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_ordered_children_in_enhanced_meshing_workflow(new_meshing_session):
    watertight = new_meshing_session.watertight()
    assert set([repr(x) for x in watertight.tasks()]) == {
        "<Task 'Add Boundary Layers'>",
        "<Task 'Add Local Sizing'>",
        "<Task 'Apply Share Topology'>",
        "<Task 'Create Regions'>",
        "<Task 'Describe Geometry'>",
        "<Task 'Enclose Fluid Regions (Capping)'>",
        "<Task 'Generate the Surface Mesh'>",
        "<Task 'Generate the Volume Mesh'>",
        "<Task 'Import Geometry'>",
        "<Task 'Update Boundaries'>",
        "<Task 'Update Regions'>",
    }


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_attrs_in_fault_tolerant_meshing_workflow(new_meshing_session):
    # Import CAD
    import_file_name = examples.download_file(
        "exhaust_system.fmd", "pyfluent/exhaust_system"
    )

    fault_tolerant = new_meshing_session.fault_tolerant()
    assert "watertight" not in dir(fault_tolerant)

    with pytest.raises(AttributeError):
        fault_tolerant.watertight()

    fault_tolerant.import_cad_and_part_management.context.set_state(0)
    fault_tolerant.import_cad_and_part_management.create_object_per.set_state("Custom")
    fault_tolerant.import_cad_and_part_management.fmd_file_name.set_state(
        import_file_name
    )

    assert fault_tolerant.import_cad_and_part_management.fmd_file_name()
    # Reinitialize the workflow:
    fault_tolerant = new_meshing_session.fault_tolerant()
    assert not fault_tolerant.import_cad_and_part_management.fmd_file_name()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_switch_between_workflows(new_meshing_session):
    meshing = new_meshing_session

    # Initialize to watertight and store
    watertight = meshing.watertight()

    assert watertight.import_geometry.arguments()

    # Wrong Attribute
    with pytest.raises(AttributeError):
        watertight.import_cad_and_part_management.arguments()

    # Initialize to fault-tolerant and store
    fault_tolerant = meshing.fault_tolerant()

    assert fault_tolerant.import_cad_and_part_management.arguments()

    # 'import_geometry' is a watertight workflow command which is not available now
    # since we have changed to fault-tolerant in the backend.
    with pytest.raises(RuntimeError):
        watertight.import_geometry.arguments()

    # Re-initialize watertight
    watertight = meshing.watertight()

    # 'import_cad_and_part_management' is a fault-tolerant workflow command which is not
    # available now since we have changed to watertight in the backend.
    with pytest.raises(RuntimeError):
        fault_tolerant.import_cad_and_part_management.arguments()

    assert watertight.import_geometry.arguments()

    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")

    # 'import_geometry' is a watertight workflow command which is not available now
    # since we have changed to fault-tolerant in the backend.
    with pytest.raises(RuntimeError):
        watertight.import_geometry.arguments()

    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")

    # 'import_cad_and_part_management' is a fault-tolerant workflow command which is not
    # available now since we have changed to watertight in the backend.
    with pytest.raises(RuntimeError):
        fault_tolerant.import_cad_and_part_management.arguments()

    # Re-initialize fault-tolerant
    fault_tolerant = meshing.fault_tolerant()
    assert fault_tolerant.import_cad_and_part_management.arguments()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.2")
def test_new_meshing_workflow_without_dm_caching(
    disable_datamodel_cache, new_meshing_session
):
    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )

    watertight = new_meshing_session.watertight()
    watertight.import_geometry.file_name = import_file_name
    watertight.import_geometry.length_unit.set_state("in")
    watertight.import_geometry()

    watertight.add_local_sizing.add_child_to_task()
    watertight.add_local_sizing()

    watertight.create_volume_mesh()

    watertight.import_geometry.rename(new_name="import_geom_wtm")
    time.sleep(2)
    assert "import_geometry" not in watertight.task_names()
    assert "import_geom_wtm" in watertight.task_names()
    assert watertight.import_geom_wtm.arguments()

    with pytest.raises(AttributeError):
        watertight.import_geometry

    watertight.delete_tasks(list_of_tasks=["add_local_sizing"])
    time.sleep(2)
    assert "add_local_sizing" not in watertight.task_names()

    assert sorted(
        [repr(x) for x in watertight.import_geom_wtm.insertable_tasks()]
    ) == sorted(
        [
            "<Insertable 'add_local_sizing' task>",
            "<Insertable 'import_boi_geometry' task>",
            "<Insertable 'set_up_rotational_periodic_boundaries' task>",
            "<Insertable 'create_local_refinement_regions' task>",
            "<Insertable 'custom_journal_task' task>",
        ]
    )

    watertight.import_geom_wtm.insertable_tasks.add_local_sizing.insert()
    assert "add_local_sizing" in watertight.task_names()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_new_meshing_workflow_switching_without_dm_caching(
    disable_datamodel_cache, new_meshing_session
):
    watertight = new_meshing_session.watertight()

    fault_tolerant = new_meshing_session.fault_tolerant()
    with pytest.raises(RuntimeError):
        watertight.import_geometry.arguments()
    assert fault_tolerant.import_cad_and_part_management.arguments()

    watertight = new_meshing_session.watertight()
    with pytest.raises(RuntimeError):
        fault_tolerant.import_cad_and_part_management.arguments()
    assert watertight.import_geometry.arguments()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.2")
def test_new_meshing_workflow_validate_arguments(new_meshing_session):
    watertight = new_meshing_session.watertight()
    watertight.create_regions.number_of_flow_volumes = 1
    with pytest.raises(ValueError):
        watertight.create_regions.number_of_flow_volumes = 1.2
    assert watertight.create_regions.arguments()["number_of_flow_volumes"] == 1
    with pytest.raises(ValueError):
        watertight.create_regions.arguments.update_dict(
            dict(number_of_flow_volumes=1.2)
        )
    assert watertight.create_regions.arguments()["number_of_flow_volumes"] == 1

    watertight.create_regions.number_of_flow_volumes = 2
    with pytest.raises(ValueError):
        watertight.create_regions.number_of_flow_volumes = 1.2
    assert watertight.create_regions.arguments()["number_of_flow_volumes"] == 2
    with pytest.raises(ValueError):
        watertight.create_regions.arguments.update_dict(
            dict(number_of_flow_volumes=1.2)
        )
    assert watertight.create_regions.arguments()["number_of_flow_volumes"] == 2

    watertight.create_regions.number_of_flow_volumes = None
    with pytest.raises(ValueError):
        watertight.create_regions.number_of_flow_volumes = 1.2
    assert watertight.create_regions.arguments()["number_of_flow_volumes"] == 1
    with pytest.raises(ValueError):
        watertight.create_regions.arguments.update_dict(
            dict(number_of_flow_volumes=1.2)
        )
    assert watertight.create_regions.arguments()["number_of_flow_volumes"] == 1


def test_camel_to_snake_case_convertor():
    assert camel_to_snake_case("ImportGeometry") == "import_geometry"
    assert camel_to_snake_case("Prism2dPreferences") == "prism_2d_preferences"
    assert camel_to_snake_case("Abc2DDef") == "abc_2d_def"
    assert camel_to_snake_case("Abc2d") == "abc_2d"
    assert camel_to_snake_case("abc2d") == "abc2d"
    assert camel_to_snake_case("AbC2d5Cb") == "ab_c2d_5_cb"
    assert camel_to_snake_case("abC2d5Cb") == "ab_c2d_5_cb"
    assert camel_to_snake_case("abC2d5Cb555klOp") == "ab_c2d_5_cb_555kl_op"
    assert camel_to_snake_case("a") == "a"
    assert camel_to_snake_case("A") == "a"
    assert camel_to_snake_case("a5$c") == "a5$c"
    assert camel_to_snake_case("A5$C") == "a5$c"
    assert camel_to_snake_case("A5Dc$") == "a5_dc$"
    assert camel_to_snake_case("Abc2DDc$") == "abc_2d_dc$"
    assert camel_to_snake_case("A2DDc$") == "a2d_dc$"
    assert camel_to_snake_case("") == ""
    assert camel_to_snake_case("BOIZoneorLabel") == "boi_zoneor_label"
    assert camel_to_snake_case("BOIZoneOrLabel") == "boi_zone_or_label"
    assert camel_to_snake_case("NumberofLayers") == "numberof_layers"
    assert camel_to_snake_case("NumberOfLayers") == "number_of_layers"
    assert (
        camel_to_snake_case("Set_Up_Rotational_Periodic_Boundaries")
        == "set_up_rotational_periodic_boundaries"
    )


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_duplicate_tasks_in_workflow(new_meshing_session):
    # Import geometry
    meshing = new_meshing_session
    watertight = meshing.watertight()

    assert sorted(
        [repr(x) for x in watertight.import_geometry.insertable_tasks()]
    ) == sorted(
        [
            "<Insertable 'import_boi_geometry' task>",
            "<Insertable 'set_up_rotational_periodic_boundaries' task>",
            "<Insertable 'create_local_refinement_regions' task>",
            "<Insertable 'custom_journal_task' task>",
        ]
    )
    assert "add_local_sizing" in watertight.task_names()
    watertight.add_local_sizing.delete()
    assert "add_local_sizing" not in watertight.task_names()
    assert "<Insertable 'add_local_sizing' task>" in [
        repr(x) for x in watertight.import_geometry.insertable_tasks()
    ]
    watertight.import_geometry.insertable_tasks.add_local_sizing.insert()
    assert "<Insertable 'add_local_sizing' task>" not in [
        repr(x) for x in watertight.import_geometry.insertable_tasks()
    ]
    watertight.import_geometry.insertable_tasks.import_boi_geometry.insert()
    watertight.import_geometry.insertable_tasks.import_boi_geometry.insert()
    watertight.import_geometry.insertable_tasks.import_boi_geometry.insert()
    assert set(watertight.task_names()) == {
        "import_geometry",
        "create_surface_mesh",
        "describe_geometry",
        "apply_share_topology",
        "enclose_fluid_regions",
        "update_boundaries",
        "create_regions",
        "update_regions",
        "add_boundary_layer",
        "create_volume_mesh",
        "add_local_sizing",
        "import_boi_geometry",
        "import_boi_geometry_1",
        "import_boi_geometry_2",
    }
    assert watertight.import_boi_geometry_1.arguments()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_object_oriented_task_inserting_in_workflows(new_meshing_session):
    meshing = new_meshing_session
    watertight = meshing.watertight()
    assert sorted(
        [repr(x) for x in watertight.import_geometry.insertable_tasks()]
    ) == sorted(
        [
            "<Insertable 'import_boi_geometry' task>",
            "<Insertable 'set_up_rotational_periodic_boundaries' task>",
            "<Insertable 'create_local_refinement_regions' task>",
            "<Insertable 'custom_journal_task' task>",
        ]
    )
    assert "set_up_rotational_periodic_boundaries" not in watertight.task_names()
    watertight.import_geometry.insertable_tasks.set_up_rotational_periodic_boundaries.insert()
    assert "set_up_rotational_periodic_boundaries" in watertight.task_names()
    assert sorted(
        [repr(x) for x in watertight.import_geometry.insertable_tasks()]
    ) == sorted(
        [
            "<Insertable 'import_boi_geometry' task>",
            "<Insertable 'create_local_refinement_regions' task>",
            "<Insertable 'custom_journal_task' task>",
        ]
    )
    watertight.import_geometry.insertable_tasks.import_boi_geometry.insert()
    watertight.import_geometry.insertable_tasks.import_boi_geometry.insert()
    assert "import_boi_geometry" in watertight.task_names()
    assert "import_boi_geometry_1" in watertight.task_names()
    assert watertight.import_boi_geometry.arguments()
    assert watertight.import_boi_geometry_1.arguments()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_loaded_workflow(new_meshing_session):
    meshing = new_meshing_session
    saved_workflow_path = examples.download_file(
        "sample_watertight_workflow.wft", "pyfluent/meshing_workflows"
    )
    loaded_workflow = meshing.load_workflow(file_path=saved_workflow_path)
    assert "set_up_rotational_periodic_boundaries" in loaded_workflow.task_names()
    assert "import_boi_geometry" in loaded_workflow.task_names()
    # The below snippet is randomly failing in CI
    # assert loaded_workflow.import_boi_geometry_1.arguments()


@pytest.mark.skip("https://github.com/ansys/pyfluent/issues/3065")
@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_created_workflow(new_meshing_session):
    meshing = new_meshing_session
    created_workflow = meshing.create_workflow()

    assert sorted([repr(x) for x in created_workflow.insertable_tasks()]) == sorted(
        [
            "<Insertable 'import_geometry' task>",
            "<Insertable 'load_cad_geometry' task>",
            "<Insertable 'import_cad_and_part_management' task>",
            "<Insertable 'custom_journal_task' task>",
        ]
    )

    created_workflow.insertable_tasks()[0].insert()

    assert created_workflow.insertable_tasks() == []

    assert "<Insertable 'add_local_sizing' task>" in [
        repr(x) for x in created_workflow.import_geometry.insertable_tasks()
    ]
    created_workflow.import_geometry.insertable_tasks.add_local_sizing.insert()
    assert "<Insertable 'add_local_sizing' task>" not in [
        repr(x) for x in created_workflow.import_geometry.insertable_tasks()
    ]
    assert sorted(created_workflow.task_names()) == sorted(
        ["import_geometry", "add_local_sizing"]
    )


@pytest.fixture
def new_meshing_session2(new_meshing_session):
    return new_meshing_session


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_independent_meshing_sessions(new_meshing_session, new_meshing_session2):
    meshing_1 = new_meshing_session
    meshing_2 = new_meshing_session2

    watertight = meshing_1.watertight()
    assert watertight.import_geometry.arguments()

    ft = meshing_1.fault_tolerant()
    assert ft.import_cad_and_part_management.arguments()

    watertight = meshing_1.watertight()
    assert watertight.import_geometry.arguments()

    fault_tolerant = meshing_2.fault_tolerant()
    assert fault_tolerant.import_cad_and_part_management.arguments()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_independent_meshing_sessions_without_dm_caching(
    disable_datamodel_cache, new_meshing_session, new_meshing_session2
):
    meshing_1 = new_meshing_session
    meshing_2 = new_meshing_session2

    watertight = meshing_1.watertight()
    assert watertight.import_geometry.arguments()

    fault_tolerant = meshing_2.fault_tolerant()
    assert fault_tolerant.import_cad_and_part_management.arguments()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.2")
def test_switching_workflow_interface(new_meshing_session):
    wt1 = new_meshing_session.watertight()
    ft = new_meshing_session.fault_tolerant()
    tw = new_meshing_session.two_dimensional_meshing()
    cw = new_meshing_session.create_workflow()
    saved_workflow_path = examples.download_file(
        "sample_watertight_workflow.wft", "pyfluent/meshing_workflows"
    )
    lw = new_meshing_session.load_workflow(file_path=saved_workflow_path)
    wt2 = new_meshing_session.watertight()
    del wt1, ft, tw, cw, lw, wt2


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_duplicate_children_of_compound_task(
    new_meshing_session, mixing_elbow_geometry_filename
):
    watertight = new_meshing_session.watertight()
    watertight.import_geometry.file_name = mixing_elbow_geometry_filename
    watertight.import_geometry()
    watertight.add_local_sizing.add_child_and_update(
        state={
            "boi_control_name": "wall",
            "boi_face_label_list": ["wall-elbow", "wall-inlet"],
            "boi_size": 10,
        }
    )
    watertight.add_local_sizing.add_child_and_update(
        state={
            "boi_control_name": "inlet",
            "boi_face_label_list": ["wall-inlet", "cold-inlet", "hot-inlet"],
            "boi_size": 10,
        }
    )
    watertight.add_local_sizing.add_child_and_update(
        state={
            "boi_control_name": "outlet",
            "boi_face_label_list": ["outlet"],
            "boi_size": 10,
        }
    )

    assert {
        "add_local_sizing_child_1",
        "add_local_sizing_child_2",
        "add_local_sizing_child_3",
    }.issubset(set(watertight.task_names()))

    assert watertight.add_local_sizing.task_names() == [
        "add_local_sizing_child_1",
        "add_local_sizing_child_2",
        "add_local_sizing_child_3",
    ]

    assert watertight.tasks()[-2].name() == "inlet"
    assert watertight.tasks()[-2].python_name() == "add_local_sizing_child_2"

    assert watertight.add_local_sizing.tasks()[-1].name() == "outlet"
    assert (
        watertight.add_local_sizing.tasks()[-1].python_name()
        == "add_local_sizing_child_3"
    )


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_current_workflow(new_meshing_session):
    meshing = new_meshing_session

    with pytest.raises(RuntimeError):
        meshing.current_workflow

    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")

    assert meshing.current_workflow.import_geometry

    with pytest.raises(AttributeError):
        meshing.current_workflow.import_cad_and_part_management

    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")

    assert meshing.current_workflow.import_cad_and_part_management

    with pytest.raises(AttributeError):
        meshing.current_workflow.import_geometry


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.1")
def test_mark_as_updated(new_meshing_session):
    meshing = new_meshing_session

    watertight = meshing.watertight()

    assert meshing.workflow.TaskObject["Import Geometry"].State() == "Out-of-date"
    assert meshing.workflow.TaskObject["Describe Geometry"].State() == "Out-of-date"
    assert meshing.workflow.TaskObject["Add Local Sizing"].State() == "Out-of-date"

    watertight.import_geometry.mark_as_updated()
    watertight.describe_geometry.mark_as_updated()
    watertight.add_local_sizing.mark_as_updated()

    assert meshing.workflow.TaskObject["Import Geometry"].State() == "Forced-up-to-date"
    assert (
        meshing.workflow.TaskObject["Describe Geometry"].State() == "Forced-up-to-date"
    )
    assert (
        meshing.workflow.TaskObject["Add Local Sizing"].State() == "Forced-up-to-date"
    )


@pytest.mark.fluent_version(">=24.1")
@pytest.mark.codegen_required
def test_accessors_for_argument_sub_items(new_meshing_session):
    meshing = new_meshing_session
    watertight = meshing.watertight()

    import_geom = watertight.import_geometry
    assert import_geom.length_unit.default_value() == "mm"
    assert "allowed_values" in dir(import_geom.length_unit)
    assert import_geom.arguments.length_unit.allowed_values() == [
        "m",
        "cm",
        "mm",
        "in",
        "ft",
        "um",
        "nm",
    ]
    assert import_geom.arguments.length_unit() == "mm"
    import_geom.length_unit.set_state("cm")
    assert import_geom.arguments.length_unit.get_state() == "cm"
    import_geom.arguments.length_unit = "in"
    assert import_geom.arguments.length_unit() == "in"
    import_geom.arguments["length_unit"] = "m"
    assert import_geom.arguments["length_unit"] == "m"
    meshing.workflow.TaskObject["Import Geometry"].Arguments = dict(LengthUnit="in")
    assert import_geom.arguments.length_unit() == "in"

    assert not import_geom.arguments.mesh_unit.is_read_only()
    assert import_geom.arguments.length_unit.is_active()
    assert not import_geom.arguments.file_name.is_read_only()
    assert not import_geom.arguments.file_name()
    import_geom.arguments.file_name = "xyz.txt"
    assert import_geom.arguments.file_name() == "xyz.txt"
    with pytest.raises(AttributeError) as msg:
        import_geom.arguments.file = "sample.txt"
    assert msg.value.args[0] == "No attribute named 'file' in 'Import Geometry'."
    with pytest.raises(AttributeError):
        import_geom.arguments.CadImportOptions.OneZonePer = "face"

    assert import_geom.arguments.cad_import_options()
    assert import_geom.arguments.cad_import_options.one_zone_per()

    assert import_geom.arguments.file_format.get_attrib_value("allowedValues") == [
        "CAD",
        "Mesh",
    ]
    assert import_geom.arguments.file_format.allowed_values() == ["CAD", "Mesh"]

    assert not import_geom.arguments.cad_import_options.one_zone_per.is_read_only()
    assert import_geom.arguments.cad_import_options.one_zone_per() == "body"
    import_geom.arguments.cad_import_options.one_zone_per.set_state("face")
    assert import_geom.arguments.cad_import_options.one_zone_per() == "face"
    import_geom.arguments.cad_import_options.one_zone_per = "object"
    assert import_geom.arguments.cad_import_options.one_zone_per() == "object"

    volume_mesh_gen = watertight.create_volume_mesh
    assert (
        volume_mesh_gen.arguments.volume_fill_controls.type.default_value()
        == "Cartesian"
    )

    # Test particular to string type (allowed_values() only available in string types)
    assert volume_mesh_gen.arguments.volume_fill_controls.type.allowed_values() == [
        "Octree",
        "Cartesian",
    ]
    feat_angle = import_geom.arguments.cad_import_options.feature_angle
    assert feat_angle.default_value() == 40.0

    # Test particular to numerical type (min() only available in numerical types)
    assert feat_angle.min() == 0.0

    # Test intended to fail in numerical type (allowed_values() only available in string types)
    with pytest.raises(AttributeError) as msg:
        assert feat_angle.allowed_values()
    assert (
        msg.value.args[0]
        == "'PyNumericalCommandArgumentsSubItem' object has no attribute 'allowed_values'"
    )

    # Test intended to fail in numerical type (allowed_values() only available in string types)
    with pytest.raises(AttributeError) as msg:
        assert import_geom.arguments.num_parts.allowed_values()
    assert (
        msg.value.args[0]
        == "'PyNumericalCommandArgumentsSubItem' object has no attribute 'allowed_values'"
    )

    # Test intended to fail in string type (min() only available in numerical types)
    with pytest.raises(AttributeError) as msg:
        assert import_geom.arguments.length_unit.min()
    assert (
        msg.value.args[0]
        == "'PyTextualCommandArgumentsSubItem' object has no attribute 'min'"
    )


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=25.1")
def test_scenario_with_common_python_names_from_fdl(new_meshing_session):
    meshing = new_meshing_session

    fault_tolerant = meshing.fault_tolerant()

    # Check if all task names are unique.
    assert len(fault_tolerant.task_names()) == len(set(fault_tolerant.task_names()))

    # APIName from fdl file
    assert "create_volume_mesh" in fault_tolerant.task_names()
    assert "generate_volume_mesh" in fault_tolerant.task_names()
    assert "generate_surface_mesh" in fault_tolerant.task_names()

    watertight = meshing.watertight()
    # Check if all task names are unique.
    assert len(watertight.task_names()) == len(set(watertight.task_names()))

    two_dimensional = meshing.two_dimensional_meshing()
    # Check if all task names are unique.
    assert len(two_dimensional.task_names()) == len(set(two_dimensional.task_names()))


@pytest.mark.skip("Failing in GitHub")
@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=25.1")
def test_return_state_changes(new_meshing_session):
    meshing = new_meshing_session
    wt = meshing.watertight()

    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )

    wt.import_geometry.file_name.set_state(import_file_name)

    # trigger creation of downstream task when this task is updated:
    wt.describe_geometry.multizone.set_state("Yes")
    wt.describe_geometry()

    assert wt.add_multizone_controls


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=25.1")
def test_recursive_update_dict(new_meshing_session):
    meshing = new_meshing_session
    fault_tolerant = meshing.fault_tolerant()
    import_file_name = examples.download_file(
        "exhaust_system.fmd", "pyfluent/exhaust_system"
    )

    import_cad = fault_tolerant.import_cad_and_part_management
    import_cad.feature_angle = 35
    import_cad.fmd_file_name = import_file_name
    import_cad()

    descr_geom = fault_tolerant.describe_geometry_and_flow
    descr_geom.arguments()
    descr_geom.flow_type = "Internal flow through the object"
    descr_geom.add_enclosure = "Yes"
    descr_geom.close_caps = "Yes"
    descr_geom.local_refinement_regions = "Yes"
    descr_geom.describe_geometry_and_flow_options.moving_objects = "Yes"
    descr_geom.describe_geometry_and_flow_options.advanced_options = True
    descr_geom.describe_geometry_and_flow_options.porous_regions = "Yes"
    descr_geom.describe_geometry_and_flow_options.enable_overset = "Yes"
    descr_geom.describe_geometry_and_flow_options.extract_edge_features = "Yes"
    descr_geom.describe_geometry_and_flow_options.zero_thickness = "Yes"
    descr_geom.arguments()
    assert meshing.workflow.TaskObject["Describe Geometry and Flow"].Arguments()[
        "DescribeGeometryAndFlowOptions"
    ] == {
        "AdvancedOptions": True,
        "EnableOverset": "Yes",
        "ExtractEdgeFeatures": "Yes",
        "MovingObjects": "Yes",
        "PorousRegions": "Yes",
        "ZeroThickness": "Yes",
    }



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for enums used in the PyFluent."""

from enum import Enum
import os

from ansys.fluent.core.exceptions import DisallowedValuesError
from ansys.fluent.core.fluent_connection import FluentConnection
import ansys.fluent.core.launcher.error_handler as exceptions
from ansys.fluent.core.launcher.launcher_utils import is_windows
from ansys.fluent.core.session_meshing import Meshing
from ansys.fluent.core.session_pure_meshing import PureMeshing
from ansys.fluent.core.session_solver import Solver
from ansys.fluent.core.session_solver_aero import SolverAero
from ansys.fluent.core.session_solver_icing import SolverIcing
from ansys.fluent.core.utils.fluent_version import FluentVersion
import ansys.platform.instancemanagement as pypim


class FluentEnum(Enum):
    """Provides the base class for Fluent-related enums.

    Accepts lowercase member names as values and supports comparison operators.
    """

    def _get_enum_map(self):
        return {}

    def get_fluent_value(self):
        """Returns the fluent value of the enum."""
        return self._get_enum_map()[self]

    def _default(self):
        return

    @classmethod
    def _missing_(cls, value: str | int | None):
        if value is None:
            return cls._default(cls)
        for member in cls:
            if member.value == value:
                return member

        def is_int():
            for m in cls:
                return True if isinstance(m.value, int) else False

        msg = ", " if is_int() else "', '"
        msg = (
            f"{msg.join(str(member.value) for member in cls)}"
            if is_int()
            else f"'{msg.join(str(member.value) for member in cls)}'"
        )
        raise DisallowedValuesError(
            f"""The specified value: {repr(value)} """
            f"""is not a supported value of {cls.__name__}."""
            f""" The supported values are: {msg}."""
        )


class LaunchMode(FluentEnum):
    """Enumerates over supported Fluent launch modes."""

    STANDALONE = "standalone"
    PIM = "pim"
    CONTAINER = "container"
    SLURM = "slurm"

    def _get_enum_map(self):
        return {
            self.STANDALONE: 1,
            self.PIM: 2,
            self.CONTAINER: 3,
            self.SLURM: 4,
        }


class FluentMode(FluentEnum):
    """Enumerates over supported Fluent modes."""

    MESHING = "meshing"
    PURE_MESHING = "pure_meshing"
    SOLVER = "solver"
    SOLVER_ICING = "solver_icing"
    SOLVER_AERO = "solver_aero"

    def _default(self):
        return self.SOLVER

    def _get_enum_map(self):
        return {
            self.MESHING: Meshing,
            self.PURE_MESHING: PureMeshing,
            self.SOLVER: Solver,
            self.SOLVER_ICING: SolverIcing,
            self.SOLVER_AERO: SolverAero,
        }

    @staticmethod
    def is_meshing(mode: "FluentMode") -> bool:
        """Check if the current mode is meshing.

        Parameters
        ----------
        mode : FluentMode
            mode

        Returns
        -------
        bool
            ``True`` if the mode is ``FluentMode.MESHING`` or ``FluentMode.PURE_MESHING``,
            ``False`` otherwise.
        """
        return mode in [FluentMode.MESHING, FluentMode.PURE_MESHING]


class UIMode(FluentEnum):
    """Provides supported user interface mode of Fluent."""

    NO_GUI_OR_GRAPHICS = "no_gui_or_graphics"
    NO_GRAPHICS = "no_graphics"
    NO_GUI = "no_gui"
    HIDDEN_GUI = "hidden_gui"
    GUI = "gui"

    def _default(self):
        # Not using NO_GUI in windows as it opens a new cmd or
        # shows Fluent output in the current cmd if start <launch_string> is not used
        return self.HIDDEN_GUI if is_windows() else self.NO_GUI

    def _get_enum_map(self):
        return {
            self.NO_GUI_OR_GRAPHICS: ("g",),
            self.NO_GRAPHICS: ("gr",),
            self.NO_GUI: ("gu",),
            self.HIDDEN_GUI: ("hidden",),
            self.GUI: ("",),
        }


class Dimension(FluentEnum):
    """Geometric dimensionality of the Fluent simulation."""

    TWO = 2
    THREE = 3

    def _default(self):
        return self.THREE

    def _get_enum_map(self):
        return {
            self.TWO: ("2d",),
            self.THREE: ("3d",),
        }


class Precision(FluentEnum):
    """Floating point precision."""

    SINGLE = "single"
    DOUBLE = "double"

    def _default(self):
        return self.DOUBLE

    def _get_enum_map(self):
        return {
            self.SINGLE: ("",),
            self.DOUBLE: ("dp",),
        }


class FluentWindowsGraphicsDriver(FluentEnum):
    """Provides supported graphics driver of Fluent in Windows."""

    NULL = "null"
    MSW = "msw"
    DX11 = "dx11"
    OPENGL2 = "opengl2"
    OPENGL = "opengl"
    AUTO = "auto"

    def _default(self):
        return self.AUTO

    def _get_enum_map(self):
        return {
            self.NULL: ("null",),
            self.MSW: ("msw",),
            self.DX11: ("dx11",),
            self.OPENGL2: ("opengl2",),
            self.OPENGL: ("opengl",),
            self.AUTO: ("",),
        }


class FluentLinuxGraphicsDriver(FluentEnum):
    """Provides supported graphics driver of Fluent in Linux."""

    NULL = "null"
    X11 = "x11"
    OPENGL2 = "opengl2"
    OPENGL = "opengl"
    AUTO = "auto"

    def _default(self):
        return self.AUTO

    def _get_enum_map(self):
        return {
            self.NULL: ("null",),
            self.X11: ("x11",),
            self.OPENGL2: ("opengl2",),
            self.OPENGL: ("opengl",),
            self.AUTO: ("",),
        }


def _get_fluent_launch_mode(start_container, container_dict, scheduler_options):
    """Get the Fluent launch mode.

    Parameters
    ----------
    start_container: bool
        Whether to launch a Fluent Docker container image.
    container_dict: dict
        Dictionary for Fluent Docker container configuration.

    Returns
    -------
    fluent_launch_mode: LaunchMode
        Fluent launch mode.
    """
    if pypim.is_configured():
        fluent_launch_mode = LaunchMode.PIM
    elif start_container is True or (
        start_container is None
        and (container_dict or os.getenv("PYFLUENT_LAUNCH_CONTAINER") == "1")
    ):
        fluent_launch_mode = LaunchMode.CONTAINER
    # Currently, only Slurm scheduler is supported and within SlurmLauncher we check the value of the scheduler
    elif scheduler_options:
        fluent_launch_mode = LaunchMode.SLURM
    else:
        fluent_launch_mode = LaunchMode.STANDALONE
    return fluent_launch_mode


def _get_graphics_driver(
    graphics_driver: (
        FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str | None
    ) = None,
):
    if isinstance(
        graphics_driver, (FluentWindowsGraphicsDriver, FluentLinuxGraphicsDriver)
    ):
        graphics_driver = graphics_driver.value
    graphics_driver = (
        FluentWindowsGraphicsDriver(graphics_driver)
        if is_windows()
        else FluentLinuxGraphicsDriver(graphics_driver)
    )
    return graphics_driver


def _get_running_session_mode(
    fluent_connection: FluentConnection, mode: FluentMode | None = None
):
    """Get the mode of the running session if the mode has not been explicitly given."""
    if mode:
        session_mode = mode
    else:
        try:
            session_mode = fluent_connection._connection_interface.get_mode()
        except Exception as ex:
            raise exceptions.InvalidPassword() from ex
    return session_mode.get_fluent_value()


def _get_standalone_launch_fluent_version(argvals) -> FluentVersion | None:
    """Determine the Fluent version during the execution of the ``launch_fluent()``
    method in standalone mode.

    The search for the version is performed in this order:

    1. The ``product_version`` parameter passed with the ``launch_fluent`` method.
    2. The latest Ansys version from ``AWP_ROOTnnn``` environment variables.

    Returns
    -------
    FluentVersion, optional
        Fluent version or ``None``
    """

    # Look for Fluent version in the following order:
    # 1. product_version parameter passed with launch_fluent
    product_version = argvals.get("product_version")
    if product_version:
        return FluentVersion(product_version)

    # If fluent_path is provided, we cannot determine the Fluent version, so returning None.
    if argvals.get("fluent_path"):
        return None

    # (DEV) if "PYFLUENT_FLUENT_ROOT" environment variable is defined, we cannot
    # determine the Fluent version, so returning None.
    if os.getenv("PYFLUENT_FLUENT_ROOT"):
        return None

    # 2. the latest ANSYS version from AWP_ROOT environment variables
    return FluentVersion.get_latest_installed()


def _validate_gpu(gpu: bool | list, dimension: int):
    """Raise an exception if the GPU Solver is unsupported.

    Parameters
    ----------
    gpu : bool or list, optional
        This option will start Fluent with the GPU Solver.
    dimension : int, optional
        Geometric dimensionality of the Fluent simulation.
    """
    if Dimension(dimension) == Dimension.TWO and gpu:
        raise exceptions.GPUSolverSupportError()


def _get_argvals_and_session(argvals):
    _validate_gpu(argvals["gpu"], argvals["dimension"])
    argvals["graphics_driver"] = _get_graphics_driver(argvals["graphics_driver"])
    argvals["mode"] = FluentMode(argvals["mode"])
    del argvals["self"]
    new_session = argvals["mode"].get_fluent_value()
    return argvals, new_session



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Wrappers over SVAR gRPC service of Fluent."""

import math
from typing import Dict, List
import warnings

import grpc
import numpy as np

from ansys.api.fluent.v0 import field_data_pb2 as FieldDataProtoModule
from ansys.api.fluent.v0 import svar_pb2 as SvarProtoModule
from ansys.api.fluent.v0 import svar_pb2_grpc as SvarGrpcModule
from ansys.fluent.core.pyfluent_warnings import PyFluentDeprecationWarning
from ansys.fluent.core.services.field_data import (
    _FieldDataConstants,
    override_help_text,
)
from ansys.fluent.core.services.interceptors import (
    GrpcErrorInterceptor,
    TracingInterceptor,
)
from ansys.fluent.core.solver.error_message import allowed_name_error_message


class SolutionVariableService:
    """SVAR service of Fluent."""

    def __init__(self, channel: grpc.Channel, metadata):
        """__init__ method of SVAR service class."""
        intercept_channel = grpc.intercept_channel(
            channel,
            GrpcErrorInterceptor(),
            TracingInterceptor(),
        )
        self.__stub = SvarGrpcModule.svarStub(intercept_channel)
        self.__metadata = metadata

    def get_data(self, request):
        """GetSvarData RPC of SVAR service."""
        return self.__stub.GetSvarData(request, metadata=self.__metadata)

    def set_data(self, request):
        """SetSvarData RPC of SVAR service."""
        return self.__stub.SetSvarData(request, metadata=self.__metadata)

    def get_variables_info(self, request):
        """GetSvarsInfo RPC of SVAR service."""
        return self.__stub.GetSvarsInfo(request, metadata=self.__metadata)

    def get_zones_info(self, request):
        """GetZonesInfo RPC of SVAR service."""
        return self.__stub.GetZonesInfo(request, metadata=self.__metadata)


class SolutionVariableInfo:
    """Provide access to Fluent SVARs and Zones information.

    Examples
    --------
    >>> solution_variable_info = solver_session.fields.solution_variable_info
    >>> wall_fluid_info = solution_variable_info.get_variables_info(zone_names=['wall' , "fluid"], domain_name="mixture")
    >>> print(wall_fluid_info.solution_variables)
    >>> ['SV_CENTROID', 'SV_D', 'SV_H', 'SV_K', 'SV_P', 'SV_T', 'SV_U', 'SV_V', 'SV_W']
    >>> solution_variable_info_centroid = wall_fluid_info['SV_CENTROID']
    >>> print(solution_variable_info_centroid)
    >>> name:SV_CENTROID dimension:3 field_type:<class 'numpy.float64'>
    >>> zones_info = solution_variable_info.get_zones_info()
    >>> print(zones_info.zones)
    >>> ['fluid', 'wall', 'symmetry', 'pressure-outlet-7', 'velocity-inlet-6', 'velocity-inlet-5', 'default-interior']
    >>> zone_info = zones_info['wall']
    >>> print(zone_info)
    >>> name:wall count: 3630 zone_id:3 zone_type:wall thread_type:Face
    """

    class SolutionVariables:
        """Class containing information for multiple solution variables."""

        class SolutionVariable:
            """Class containing information for single solution variable."""

            def __init__(self, solution_variable_info):
                """Initialize SolutionVariable."""
                self.name = solution_variable_info.name
                self.dimension = solution_variable_info.dimension
                self.field_type = _FieldDataConstants.proto_field_type_to_np_data_type[
                    solution_variable_info.fieldType
                ]

            def __repr__(self):
                return f"name:{self.name} dimension:{self.dimension} field_type:{self.field_type}"

        def __init__(self, solution_variables_info):
            """Initialize SolutionVariables."""
            self._solution_variables_info = {}
            for solution_variable_info in solution_variables_info:
                self._solution_variables_info[solution_variable_info.name] = (
                    SolutionVariableInfo.SolutionVariables.SolutionVariable(
                        solution_variable_info
                    )
                )

        def _filter(self, solution_variables_info):
            self._solution_variables_info = {
                k: v
                for k, v in self._solution_variables_info.items()
                if k
                in [
                    solution_variable_info.name
                    for solution_variable_info in solution_variables_info
                ]
            }

        def __getitem__(self, name):
            return self._solution_variables_info.get(name, None)

        @property
        def solution_variables(self) -> List[str]:
            """Solution variables."""
            return list(self._solution_variables_info.keys())

        @property
        def svars(self) -> List[str]:
            """Solution variables."""
            warnings.warn(
                "svars is deprecated, use solution_variables instead",
                PyFluentDeprecationWarning,
            )
            return self.solution_variables

    class ZonesInfo:
        """Class containing information for multiple zones."""

        class ZoneInfo:
            """Class containing information for single zone."""

            class PartitionsInfo:
                """Class containing information for partitions."""

                def __init__(self, partition_info):
                    """Initialize PartitionsInfo."""
                    self.count = partition_info.count
                    self.start_index = (
                        partition_info.startIndex if self.count > 0 else 0
                    )
                    self.end_index = partition_info.endIndex if self.count > 0 else 0

            def __init__(self, zone_info):
                """Initialize ZoneInfo."""
                self.name = zone_info.name
                self.zone_id = zone_info.zoneId
                self.zone_type = zone_info.zoneType
                self.thread_type = zone_info.threadType
                self.partitions_info = [
                    self.PartitionsInfo(partition_info)
                    for partition_info in zone_info.partitionsInfo
                ]

            @property
            def count(self) -> int:
                """Get zone count."""
                return sum(
                    [partition_info.count for partition_info in self.partitions_info]
                )

            def __repr__(self):
                partition_str = ""
                for i, partition_info in enumerate(self.partitions_info):
                    partition_str += f"\n\t{i}. {partition_info.count}[{partition_info.start_index}:{partition_info.end_index}]"
                return f"name:{self.name} count: {self.count} zone_id:{self.zone_id} zone_type:{self.zone_type} threadType:{'Cell' if self.thread_type == SvarProtoModule.ThreadType.CELL_THREAD else 'Face'}{partition_str}"

        def __init__(self, zones_info, domains_info):
            """Initialize ZonesInfo."""
            self._zones_info = {}
            self._domains_info = {}
            for zone_info in zones_info:
                self._zones_info[zone_info.name] = self.ZoneInfo(zone_info)
            for domain_info in domains_info:
                self._domains_info[domain_info.name] = domain_info.domainId

        def __getitem__(self, name):
            return self._zones_info.get(name, None)

        @property
        def zones(self) -> List[str]:
            """Get zone names."""
            return list(self._zones_info.keys())

        @property
        def domains(self) -> List[str]:
            """Get domain names."""
            return list(self._domains_info.keys())

        def domain_id(self, domain_name) -> int:
            """Get domain id."""
            return self._domains_info.get(domain_name, None)

    def __init__(
        self,
        service: SolutionVariableService,
    ):
        """Initialize SolutionVariableInfo."""
        self._service = service

    def get_variables_info(
        self, zone_names: List[str], domain_name: str | None = "mixture"
    ) -> SolutionVariables:
        """Get SVARs info for zones in the domain.

        Parameters
        ----------
        zone_names : List[str]
            List of zone names.
        domain_name: str, optional
            Domain name.The default is ``mixture``.

        Returns
        -------
        SolutionVariableInfo.SolutionVariables
            Object containing information for SVARs which are common for list of zone names.
        """

        allowed_zone_names = _AllowedZoneNames(self)
        allowed_domain_names = _AllowedDomainNames(self)
        solution_variables_info = None
        for zone_name in zone_names:
            request = SvarProtoModule.GetSvarsInfoRequest(
                domainId=allowed_domain_names.valid_name(domain_name),
                zoneId=allowed_zone_names.valid_name(zone_name),
            )
            response = self._service.get_variables_info(request)
            if solution_variables_info is None:
                solution_variables_info = SolutionVariableInfo.SolutionVariables(
                    response.svarsInfo
                )
            else:
                solution_variables_info._filter(response.svarsInfo)
        return solution_variables_info

    def get_svars_info(
        self, zone_names: List[str], domain_name: str | None = "mixture"
    ) -> SolutionVariables:
        """Get solution variables info."""
        warnings.warn(
            "get_svars_info is deprecated, use get_variables_info instead",
            PyFluentDeprecationWarning,
        )
        return self.get_variables_info(zone_names=zone_names, domain_name=domain_name)

    def get_zones_info(self) -> ZonesInfo:
        """Get Zones info.

        Parameters
        ----------
        None

        Returns
        -------
        SolutionVariableInfo.ZonesInfo
            Object containing information for all zones.
        """
        request = SvarProtoModule.GetZonesInfoRequest()
        response = self._service.get_zones_info(request)
        return SolutionVariableInfo.ZonesInfo(response.zonesInfo, response.domainsInfo)


class SvarError(ValueError):
    """Exception class for errors in solution variable name."""

    def __init__(self, solution_variable_name: str, allowed_values: List[str]):
        """Initialize SvarError."""
        self.solution_variable_name = solution_variable_name
        super().__init__(
            allowed_name_error_message(
                context="solution variable",
                trial_name=solution_variable_name,
                allowed_values=allowed_values,
            )
        )


class ZoneError(ValueError):
    """Exception class for errors in Zone name."""

    def __init__(self, zone_name: str, allowed_values: List[str]):
        """Initialize ZoneError."""
        self.zone_name = zone_name
        super().__init__(
            allowed_name_error_message(
                context="zone", trial_name=zone_name, allowed_values=allowed_values
            )
        )


class _AllowedNames:
    def is_valid(self, name):
        """Check whether a given name is valid or not."""
        return name in self()


class _AllowedSvarNames:
    def __init__(self, solution_variable_info: SolutionVariableInfo):
        self._solution_variable_info = solution_variable_info

    def __call__(
        self, zone_names: List[str], domain_name: str | None = "mixture"
    ) -> List[str]:
        return self._solution_variable_info.get_variables_info(
            zone_names=zone_names, domain_name=domain_name
        ).solution_variables

    def is_valid(
        self,
        solution_variable_name,
        zone_names: List[str],
        domain_name: str | None = "mixture",
    ):
        """Check whether solution variable name is valid or not."""
        return solution_variable_name in self(
            zone_names=zone_names, domain_name=domain_name
        )

    def valid_name(
        self,
        solution_variable_name,
        zone_names: List[str],
        domain_name: str | None = "mixture",
    ):
        """Get a valid solution variable name.

        Raises
        ------
        SvarError
            If the given solution variable name is invalid.
        """
        if not self.is_valid(
            solution_variable_name, zone_names=zone_names, domain_name=domain_name
        ):
            raise SvarError(
                solution_variable_name=solution_variable_name,
                allowed_values=self(zone_names=zone_names, domain_name=domain_name),
            )
        return solution_variable_name


class _AllowedZoneNames(_AllowedNames):
    def __init__(self, solution_variable_info: SolutionVariableInfo):
        self._zones_info = solution_variable_info.get_zones_info()

    def __call__(self) -> List[str]:
        return self._zones_info.zones

    def valid_name(self, zone_name):
        """Get a valid zone name.

        Raises
        ------
        ZoneError
            If the given zone name is invalid.
        """
        if not self.is_valid(zone_name):
            raise ZoneError(
                zone_name=zone_name,
                allowed_values=self(),
            )
        return self._zones_info[zone_name].zone_id


class _AllowedDomainNames(_AllowedNames):
    def __init__(self, solution_variable_info: SolutionVariableInfo):
        self._zones_info = solution_variable_info.get_zones_info()

    def __call__(self) -> List[str]:
        return self._zones_info.domains

    def valid_name(self, domain_name):
        """Get a valid domain name.

        Raises
        ------
        ZoneError
            If the given domain name is invalid.
        """
        if not self.is_valid(domain_name):
            raise ZoneError(
                domain_name=domain_name,
                allowed_values=self(),
            )
        return self._zones_info.domain_id(domain_name)


class _SvarMethod:
    class _Arg:
        def __init__(self, accessor):
            self._accessor = accessor

        def allowed_values(self):
            """Get allowed values."""
            return sorted(self._accessor())

    def __init__(self, svar_accessor, args_allowed_values_accessors):
        self._svar_accessor = svar_accessor
        for arg_name, accessor in args_allowed_values_accessors.items():
            setattr(self, arg_name, _SvarMethod._Arg(accessor))

    def __call__(self, *args, **kwargs):
        return self._svar_accessor(*args, **kwargs)


def extract_svars(solution_variables_data):
    """Extracts SVAR data via a server call."""

    def _extract_svar(field_datatype, field_size, solution_variables_data):
        field_arr = np.empty(field_size, dtype=field_datatype)
        field_datatype_item_size = np.dtype(field_datatype).itemsize
        index = 0
        for solution_variable_data in solution_variables_data:
            chunk = solution_variable_data.payload
            if chunk.bytePayload:
                count = min(
                    len(chunk.bytePayload) // field_datatype_item_size,
                    field_size - index,
                )
                field_arr[index : index + count] = np.frombuffer(
                    chunk.bytePayload, field_datatype, count=count
                )
                index += count
                if index == field_size:
                    return field_arr
            else:
                payload = (
                    chunk.floatPayload.payload
                    or chunk.intPayload.payload
                    or chunk.doublePayload.payload
                    or chunk.longPayload.payload
                )
                count = len(payload)
                field_arr[index : index + count] = np.fromiter(
                    payload, dtype=field_datatype
                )
                index += count
                if index == field_size:
                    return field_arr

    zones_svar_data = {}
    for array in solution_variables_data:
        if array.WhichOneof("array") == "payloadInfo":
            zones_svar_data[array.payloadInfo.zone] = _extract_svar(
                _FieldDataConstants.proto_field_type_to_np_data_type[
                    array.payloadInfo.fieldType
                ],
                array.payloadInfo.fieldSize,
                solution_variables_data,
            )
        elif array.WhichOneof("array") == "header":
            continue

    return zones_svar_data


class SolutionVariableData:
    """Provides access to Fluent SVAR data on zones.

    Examples
    --------
    >>> solution_variable_data = solver_session.fields.solution_variable_data
    >>> sv_t_wall_fluid=solver_session.fields.solution_variable_data.get_data(solution_variable_name="SV_T", domain_name="mixture", zone_names=["fluid", "wall"])
    >>> print(sv_t_wall_fluid.domain)
    >>> 'mixture'
    >>> print(sv_t_wall_fluid.zones)
    >>> ['fluid', 'wall']
    >>> fluid_temp = sv_t_wall_fluid['fluid']
    >>> print(fluid_temp.size)
    >>> 13852
    >>> print(fluid_temp.dtype)
    >>> float64
    >>> print(fluid_temp)
    >>> array([600., 600., 600., ..., 600., 600., 600.])
    >>> wall_temp_array = solution_variable_data.create_empty_array("SV_T", "wall")
    >>> fluid_temp_array =solution_variable_data.create_empty_array("SV_T", "fluid")
    >>> wall_temp_array[:]= 500
    >>> fluid_temp_array[:]= 600
    >>> zone_names_to_solution_variable_data = {'wall':wall_temp_array, 'fluid':fluid_temp_array}
    >>> solution_variable_data.set_data(solution_variable_name="SV_T", domain_name="mixture", zone_names_to_solution_variable_data=zone_names_to_solution_variable_data)
    """

    class Data:
        """Solution variable data."""

        def __init__(self, domain_name, zone_id_name_map, solution_variable_data):
            """Initialize Data."""
            self._domain_name = domain_name
            self._data = {
                zone_id_name_map[zone_id]: zone_data
                for zone_id, zone_data in solution_variable_data.items()
            }

        @property
        def domain(self):
            """Domain name."""
            return self._domain_name

        @property
        def zones(self):
            """Zone name."""
            return list(self._data.keys())

        @property
        def data(self):
            """Solution variable data."""
            return self._data

        def __getitem__(self, name):
            return self._data.get(name, None)

    def __init__(
        self,
        service: SolutionVariableService,
        solution_variable_info: SolutionVariableInfo,
    ):
        """Initialize SolutionVariableData."""
        self._service = service
        self._solution_variable_info = solution_variable_info

        self.get_data = override_help_text(
            _SvarMethod(
                svar_accessor=self.get_data,
                args_allowed_values_accessors={},
            ),
            SolutionVariableData.get_data,
        )

    def _update_solution_variable_info(self):
        self._allowed_zone_names = _AllowedZoneNames(self._solution_variable_info)

        self._allowed_domain_names = _AllowedDomainNames(self._solution_variable_info)

        self._allowed_solution_variable_names = _AllowedSvarNames(
            self._solution_variable_info
        )

    def create_empty_array(
        self,
        solution_variable_name: str,
        zone_name: str,
        domain_name: str | None = "mixture",
    ) -> np.zeros:
        """Get numpy zeros array for the SVAR on a zone.

        This array can be populated  with values to set SVAR data.
        """
        self._update_solution_variable_info()

        zones_info = self._solution_variable_info.get_zones_info()
        if zone_name in zones_info.zones:
            solution_variables_info = self._solution_variable_info.get_variables_info(
                zone_names=[zone_name], domain_name=domain_name
            )
            if solution_variable_name in solution_variables_info.solution_variables:
                return np.zeros(
                    zones_info[zone_name].count
                    * solution_variables_info[solution_variable_name].dimension,
                    dtype=solution_variables_info[solution_variable_name].field_type,
                )

    def get_data(
        self,
        solution_variable_name: str,
        zone_names: List[str],
        domain_name: str | None = "mixture",
    ) -> Data:
        """Get SVAR data on zones.

        Parameters
        ----------
        solution_variable_name : str
            Name of the SVAR.
        zone_names: List[str]
            Zone names list for SVAR data.
        domain_name : str, optional
            Domain name. The default is ``mixture``.

        Returns
        -------
        SolutionVariableData.Data
            Object containing SVAR data.
        """
        self._update_solution_variable_info()
        svars_request = SvarProtoModule.GetSvarDataRequest(
            provideBytesStream=_FieldDataConstants.bytes_stream,
            chunkSize=_FieldDataConstants.chunk_size,
        )
        svars_request.domainId = self._allowed_domain_names.valid_name(domain_name)
        svars_request.name = self._allowed_solution_variable_names.valid_name(
            solution_variable_name, zone_names, domain_name
        )
        zone_id_name_map = {}
        for zone_name in zone_names:
            zone_id = self._allowed_zone_names.valid_name(zone_name)
            zone_id_name_map[zone_id] = zone_name
            svars_request.zones.append(zone_id)

        return SolutionVariableData.Data(
            domain_name,
            zone_id_name_map,
            extract_svars(self._service.get_data(svars_request)),
        )

    def get_svar_data(
        self,
        svar_name: str,
        zone_names: List[str],
        domain_name: str | None = "mixture",
    ) -> Data:
        """Get solution variable data."""
        warnings.warn(
            "get_svar_data is deprecated, use get_data instead",
            PyFluentDeprecationWarning,
        )
        return self.get_data(
            solution_variable_name=svar_name,
            zone_names=zone_names,
            domain_name=domain_name,
        )

    def set_data(
        self,
        solution_variable_name: str,
        zone_names_to_solution_variable_data: Dict[str, np.array],
        domain_name: str | None = "mixture",
    ) -> None:
        """Set SVAR data on zones.

        Parameters
        ----------
        solution_variable_name : str
            Name of the SVAR.
        zone_names_to_solution_variable_data: Dict[str, np.array]
            Dictionary containing zone names for SVAR data.
        domain_name : str, optional
            Domain name. The default is ``mixture``.

        Returns
        -------
        None
        """
        self._update_solution_variable_info()
        domain_id = self._allowed_domain_names.valid_name(domain_name)
        zone_ids_to_svar_data = {
            self._allowed_zone_names.valid_name(zone_name): solution_variable_data
            for zone_name, solution_variable_data in zone_names_to_solution_variable_data.items()
        }

        def generate_set_data_requests():
            set_data_requests = []

            set_data_requests.append(
                SvarProtoModule.SetSvarDataRequest(
                    header=SvarProtoModule.SvarHeader(
                        name=solution_variable_name, domainId=domain_id
                    )
                )
            )

            for zone_id, solution_variable_data in zone_ids_to_svar_data.items():
                max_array_size = (
                    _FieldDataConstants.chunk_size
                    / np.dtype(solution_variable_data.dtype).itemsize
                )
                solution_variable_data_list = np.array_split(
                    solution_variable_data,
                    math.ceil(solution_variable_data.size / max_array_size),
                )
                set_data_requests.append(
                    SvarProtoModule.SetSvarDataRequest(
                        payloadInfo=SvarProtoModule.Info(
                            fieldType=_FieldDataConstants.np_data_type_to_proto_field_type[
                                solution_variable_data.dtype.type
                            ],
                            fieldSize=solution_variable_data.size,
                            zone=zone_id,
                        )
                    )
                )
                set_data_requests += [
                    SvarProtoModule.SetSvarDataRequest(
                        payload=(
                            SvarProtoModule.Payload(
                                floatPayload=FieldDataProtoModule.FloatPayload(
                                    payload=solution_variable_data
                                )
                            )
                            if solution_variable_data.dtype.type == np.float32
                            else (
                                SvarProtoModule.Payload(
                                    doublePayload=FieldDataProtoModule.DoublePayload(
                                        payload=solution_variable_data
                                    )
                                )
                                if solution_variable_data.dtype.type == np.float64
                                else (
                                    SvarProtoModule.Payload(
                                        intPayload=FieldDataProtoModule.IntPayload(
                                            payload=solution_variable_data
                                        )
                                    )
                                    if solution_variable_data.dtype.type == np.int32
                                    else SvarProtoModule.Payload(
                                        longPayload=FieldDataProtoModule.LongPayload(
                                            payload=solution_variable_data
                                        )
                                    )
                                )
                            )
                        )
                    )
                    for solution_variable_data in solution_variable_data_list
                    if solution_variable_data.size > 0
                ]

            for set_data_request in set_data_requests:
                yield set_data_request

        self._service.set_data(generate_set_data_requests())

    def set_svar_data(
        self,
        svar_name: str,
        zone_names_to_svar_data: List[str],
        domain_name: str | None = "mixture",
    ) -> Data:
        """Set solution variable data."""
        warnings.warn(
            "set_svar_data is deprecated, use set_data instead",
            PyFluentDeprecationWarning,
        )
        return self.set_data(
            solution_variable_name=svar_name,
            zone_names_to_solution_variable_data=zone_names_to_svar_data,
            domain_name=domain_name,
        )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module to fix text documentation for a data object."""

from io import StringIO


def escape_wildcards(doc: str):
    """Escape wildcards."""
    new_doc = StringIO()
    prev_c = None
    for i, c in enumerate(doc):
        if c == "*":
            if prev_c == "\\":
                new_doc.write(r"\*")
            else:
                new_doc.write(r"\\*")
        else:
            new_doc.write(c)
        prev_c = c
    return new_doc.getvalue()


def fix_definition_list_in_class_doc(doc: str):
    """Fix definition list in class docstring."""
    old_lines = doc.splitlines(keepends=True)
    new_lines = []
    for i, line in enumerate(old_lines):
        if line.strip().startswith("-") and not line.strip().startswith("--"):
            if i == 0:
                new_lines.append("\n")
                new_lines.append(line)
            elif i == len(old_lines) - 1:
                new_lines.append(line)
                new_lines.append("\n")
            else:
                prev_line = old_lines[i - 1].strip()
                if prev_line and not prev_line.startswith("-"):
                    new_lines.append("\n")
                new_lines.append(line)
                next_line = old_lines[i + 1].strip()
                if next_line and not next_line.startswith("-"):
                    new_lines.append("\n")
        else:
            new_lines.append(line)
    return "".join(new_lines)


def fix_settings_doc(doc: str):
    """Fix settings docstring."""
    doc = escape_wildcards(doc)
    return fix_definition_list_in_class_doc(doc)



"""Provides a module to copy files from the Ansys installation directory."""

from pathlib import Path
import shutil
import sys


def create_file_folders_list(files_list: list, fluent_version: Path | str):
    """Create a list of files and folders specified in a text file.

    Parameters
    ----------
    files_list: list
        List of text files containing relative paths of files and folders.
    fluent_version: Path | str
        Path of ``docker/fluent_<version>`` folder.

    Returns
    -------
    file_folders: list
        List of files and folders specified in given text files.
    """
    file_folders = []
    for file in files_list:
        with open(Path(fluent_version) / file, "r") as f:
            lines = f.readlines()
            file_folders.extend([line.rstrip("\n") for line in lines])
    return file_folders


def copy_files(src: Path | str, fluent_version: Path | str):
    """Copy files from the Ansys installation directory.

    Parameters
    ----------
    src: Path | str
        Path of ``ansys_inc`` folder in the Ansys installation directory.
    fluent_version: Path | str
        Path of ``docker/fluent_<version>`` folder.
    """
    copy_files = ["cadList.txt", "ceiList.txt", "cfdpostList.txt", "fluentList.txt"]
    remove_files = ["excludeCEIList.txt", "excludeFluentList.txt"]
    copy_list = create_file_folders_list(
        files_list=copy_files, fluent_version=fluent_version
    )
    remove_list = create_file_folders_list(
        files_list=remove_files, fluent_version=fluent_version
    )
    dst = Path(fluent_version) / "ansys_inc"
    for file in copy_list:
        source = Path(src) / file
        destination = Path(dst) / file
        if Path(source).is_file():
            Path(destination).parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(src=source, dst=destination)
        elif Path(source).is_dir():
            Path(destination).mkdir(parents=True, exist_ok=True)
            shutil.copytree(src=source, dst=destination, dirs_exist_ok=True)
    for file in remove_list:
        destination = Path(dst) / file
        if Path(destination).is_file():
            Path(destination).unlink()
        elif Path(destination).is_dir():
            shutil.rmtree(Path(destination))


if __name__ == "__main__":
    copy_files(src=sys.argv[1], fluent_version=sys.argv[2])



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest


@pytest.mark.settings_only
@pytest.mark.fluent_version(">=24.1")
def test_expression(mixing_elbow_settings_session):
    solver_session = mixing_elbow_settings_session
    # Case file already has energy model turned on
    # solver_session.setup.models.energy.enabled = True
    expressions = solver_session.setup.named_expressions
    expressions["r"] = {}
    expressions["r"] = {"definition": "(Position.z**2.0 +Position.x**2.0)**0.5"}
    expressions["r1"] = {}
    expressions["r1"] = {"definition": "1-(r/0.014[m])"}
    expressions["v1"] = {}
    expressions["v1"] = {"definition": "r1**(1.0/6.0)"}
    expressions["vel_cold"] = {}
    expressions["vel_cold"] = {"definition": "1.264 * 1.43 [m s^-1] * max(0,v1)"}

    assert expressions["r"].definition() == "(Position.z**2.0 +Position.x**2.0)**0.5"
    assert expressions["r1"].definition() == "1-(r/0.014[m])"
    assert expressions["v1"].definition() == "r1**(1.0/6.0)"
    assert expressions["vel_cold"].definition() == "1.264 * 1.43 [m s^-1] * max(0,v1)"

    velocity_inlet = solver_session.setup.boundary_conditions.velocity_inlet
    velocity_inlet["cold-inlet"].momentum = {"velocity": "vel_cold"}
    velocity_inlet["hot-inlet"].momentum = {"velocity": "max(vel_cold, 1.5 [m/s])"}
    assert velocity_inlet["cold-inlet"].momentum.velocity() == {
        "option": "value",
        "value": "vel_cold",
    }
    assert velocity_inlet["hot-inlet"].momentum.velocity() == {
        "option": "value",
        "value": "max(vel_cold, 1.5 [m/s])",
    }



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for launching Fluent in container mode.

Examples
--------

>>> from ansys.fluent.core.launcher.launcher import create_launcher
>>> from ansys.fluent.core.launcher.pyfluent_enums import LaunchMode, FluentMode

>>> container_meshing_launcher = create_launcher(LaunchMode.CONTAINER, mode=FluentMode.MESHING)
>>> container_meshing_session = container_meshing_launcher()

>>> container_solver_launcher = create_launcher(LaunchMode.CONTAINER)
>>> container_solver_session = container_solver_launcher()
"""

import inspect
import logging
import os
from typing import Any

from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.launcher.fluent_container import (
    configure_container_dict,
    start_fluent_container,
)
from ansys.fluent.core.launcher.process_launch_string import (
    _build_fluent_launch_args_string,
)
from ansys.fluent.core.launcher.pyfluent_enums import (
    Dimension,
    FluentLinuxGraphicsDriver,
    FluentMode,
    FluentWindowsGraphicsDriver,
    Precision,
    UIMode,
    _get_argvals_and_session,
)
import ansys.fluent.core.launcher.watchdog as watchdog
from ansys.fluent.core.utils.fluent_version import FluentVersion

_THIS_DIR = os.path.dirname(__file__)
_OPTIONS_FILE = os.path.join(_THIS_DIR, "fluent_launcher_options.json")
logger = logging.getLogger("pyfluent.launcher")


class DockerLauncher:
    """Instantiates Fluent session in container mode."""

    def __init__(
        self,
        mode: FluentMode | str | None = None,
        ui_mode: UIMode | str | None = None,
        graphics_driver: (
            FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str | None
        ) = None,
        product_version: FluentVersion | str | float | int | None = None,
        dimension: Dimension | int | None = None,
        precision: Precision | str | None = None,
        processor_count: int | None = None,
        start_timeout: int = 60,
        additional_arguments: str = "",
        container_dict: dict | None = None,
        dry_run: bool = False,
        cleanup_on_exit: bool = True,
        start_transcript: bool = True,
        py: bool | None = None,
        gpu: bool | None = None,
        start_watchdog: bool | None = None,
        file_transfer_service: Any | None = None,
    ):
        """
        Launch a Fluent session in container mode.

        Parameters
        ----------
        mode : FluentMode
            Specifies the launch mode of Fluent to target a specific session type.
        ui_mode : UIMode
            Defines the user interface mode for Fluent. Options correspond to values in the ``UIMode`` enum.
        graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver
            Specifies the graphics driver for Fluent. Options are from the ``FluentWindowsGraphicsDriver`` enum
            (for Windows) or the ``FluentLinuxGraphicsDriver`` enum (for Linux).
        product_version :  FluentVersion or str or float or int, optional
            Indicates the version of Ansys Fluent to launch. For example, to use version 2025 R1, pass
            any of ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``, or ``251``. Defaults to ``None``,
            which uses the newest installed version.
        dimension : Dimension or int, optional
            Specifies the geometric dimensionality of the Fluent simulation. Defaults to ``None``,
            which corresponds to ``Dimension.THREE``. Acceptable values include ``Dimension.TWO``,
            ``Dimension.THREE``, or integers ``2`` and ``3``.
        precision : Precision or str, optional
            Defines the floating point precision. Defaults to ``None``, which corresponds to
            ``Precision.DOUBLE``. Acceptable values include ``Precision.SINGLE``,
            ``Precision.DOUBLE``, or strings ``"single"`` and ``"double"``.
        processor_count : int, optional
            Specifies the number of processors to use. Defaults to ``None``, which uses 1 processor.
            In job scheduler environments, this value limits the total number of allocated cores.
        start_timeout : int, optional
            Maximum allowable time in seconds for connecting to the Fluent server. Defaults to 60 seconds.
        additional_arguments : str, optional
            Additional command-line arguments for Fluent, formatted as they would be on the command line.
        container_dict : dict, optional
            Configuration dictionary for launching Fluent inside a Docker container. See also
            :mod:`~ansys.fluent.core.launcher.fluent_container`.
        dry_run : bool, optional
            If True, does not launch Fluent but prints configuration information instead. If dry running a
            container start, this method will return the configured ``container_dict``. Defaults to False.
        cleanup_on_exit : bool
            Determines whether to shut down the connected Fluent session upon exit or when calling
            the session's `exit()` method. Defaults to True.
        start_transcript : bool
            Indicates whether to start streaming the Fluent transcript in the client. Defaults to True;
            streaming can be controlled via `transcript.start()` and `transcript.stop()` methods on the session object.
        py : bool, optional
            If True, runs Fluent in Python mode. Defaults to None.
        gpu : bool, optional
            If True, starts Fluent with GPU Solver enabled.
        start_watchdog : bool, optional
            If True and `cleanup_on_exit` is True, an independent watchdog process is run to ensure that any local
            GUI-less Fluent sessions started by PyFluent are properly closed when the current Python process ends.
        file_transfer_service : Any, optional
            Service for uploading/downloading files to/from the server.

        Returns
        -------
        Meshing | PureMeshing | Solver | SolverIcing | dict
            Session object or configuration dictionary if ``dry_run`` is True.

        Raises
        ------
        UnexpectedKeywordArgument
            If an unexpected keyword argument is provided.

        Notes
        -----
        In job scheduler environments (e.g., SLURM, LSF, PBS), resources and compute nodes are allocated,
        and core counts are queried from these environments before being passed to Fluent.
        """
        locals_ = locals().copy()
        argvals = {
            arg: locals_.get(arg)
            for arg in inspect.getargvalues(inspect.currentframe()).args
        }
        self.argvals, self.new_session = _get_argvals_and_session(argvals)
        if self.argvals["start_timeout"] is None:
            self.argvals["start_timeout"] = 60
        self.file_transfer_service = file_transfer_service
        if self.argvals["mode"] == FluentMode.SOLVER_ICING:
            self.argvals["fluent_icing"] = True
        if self.argvals["container_dict"] is None:
            self.argvals["container_dict"] = {}
        if self.argvals["product_version"]:
            self.argvals["container_dict"][
                "image_tag"
            ] = f"v{FluentVersion(self.argvals['product_version']).value}"

        self._args = _build_fluent_launch_args_string(**self.argvals).split()
        if FluentMode.is_meshing(self.argvals["mode"]):
            self._args.append(" -meshing")

    def __call__(self):
        if self.argvals["dry_run"]:
            config_dict, *_ = configure_container_dict(
                self._args, **self.argvals["container_dict"]
            )
            from pprint import pprint

            print("\nDocker container run configuration:\n")
            print("config_dict = ")
            if os.getenv("PYFLUENT_HIDE_LOG_SECRETS") != "1":
                pprint(config_dict)
            else:
                config_dict_h = config_dict.copy()
                config_dict_h.pop("environment")
                pprint(config_dict_h)
                del config_dict_h
            return config_dict

        port, password, container = start_fluent_container(
            self._args, self.argvals["container_dict"]
        )

        fluent_connection = FluentConnection(
            port=port,
            password=password,
            file_transfer_service=self.file_transfer_service,
            cleanup_on_exit=self.argvals["cleanup_on_exit"],
            slurm_job_id=self.argvals and self.argvals.get("slurm_job_id"),
            inside_container=True,
        )

        session = self.new_session(
            fluent_connection=fluent_connection,
            scheme_eval=fluent_connection._connection_interface.scheme_eval,
            file_transfer_service=self.file_transfer_service,
            start_transcript=self.argvals["start_transcript"],
        )
        session._container = container

        if self.argvals["start_watchdog"] is None and self.argvals["cleanup_on_exit"]:
            self.argvals["start_watchdog"] = True
        if self.argvals["start_watchdog"]:
            logger.debug("Launching Watchdog for Fluent container...")
            watchdog.launch(os.getpid(), port, password)

        return session



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for post objects."""

import re

from ansys.fluent.core.solver.flunits import get_si_unit_for_fluent_quantity
from ansys.fluent.core.utils.fluent_version import FluentVersion


class IncompleteISOSurfaceDefinition(RuntimeError):
    """Raised when iso-surface definition is incomplete."""

    def __init__(self):
        """Initialize IncompleteISOSurfaceDefinition."""
        super().__init__("Iso surface definition is incomplete.")


class SurfaceCreationError(RuntimeError):
    """Raised when surface creation is unsuccessful."""

    def __init__(self):
        """Initialize SurfaceCreationError."""
        super().__init__("Surface creation is unsuccessful.")


class PostAPIHelper:
    """Class providing helper API for post objects."""

    class _SurfaceAPI:
        """Class providing APIs for surface operations."""

        def __init__(self, obj):
            self.obj = obj
            self._surface_name_on_server = self.surface_name_on_server(obj._name)

        @staticmethod
        def surface_name_on_server(local_surface_name):
            """Returns the surface name on server."""
            return "_dummy_surface_for_pyfluent:" + local_surface_name.lower()

        def _get_api_handle(self):
            return self.obj.get_root().session.results.surfaces

        def _delete_if_exist_on_server(self):
            field_info = self.obj._api_helper.field_info()
            surfaces_list = list(field_info.get_surfaces_info().keys())
            if self._surface_name_on_server in surfaces_list:
                self.delete_surface_on_server()

        def create_surface_on_server(self):
            """Creates the surface on server.

            Raises
            ------
            IncompleteISOSurfaceDefinition
                If iso-surface definition is incomplete.
            SurfaceCreationError
                If server fails to create surface.
            """
            if self.obj.definition.type() == "iso-surface":
                iso_surface = self.obj.definition.iso_surface
                field = iso_surface.field()
                iso_value = iso_surface.iso_value()
                if not field:
                    raise IncompleteISOSurfaceDefinition()
                self._delete_if_exist_on_server()
                self._get_api_handle().iso_surface[self._surface_name_on_server] = {
                    "field": field,
                    "iso_values": [iso_value],
                }
            elif self.obj.definition.type() == "plane-surface":
                plane_surface = self.obj.definition.plane_surface
                xy_plane = plane_surface.xy_plane
                yz_plane = plane_surface.yz_plane
                zx_plane = plane_surface.zx_plane
                self._delete_if_exist_on_server()
                if xy_plane():
                    method = "xy-plane"
                    position = "z"
                    value = xy_plane.z()
                elif yz_plane():
                    method = "yz-plane"
                    position = "x"
                    value = yz_plane.x()
                else:
                    method = "zx-plane"
                    position = "y"
                    value = zx_plane.y()
                self._get_api_handle().plane_surface[self._surface_name_on_server] = {
                    "method": method,
                    position: value,
                }
            field_info = self.obj._api_helper.field_info()
            surfaces_list = list(field_info.get_surfaces_info().keys())
            if self._surface_name_on_server not in surfaces_list:
                raise SurfaceCreationError()

        def delete_surface_on_server(self):
            """Deletes the surface on server."""
            if self.obj.definition.type() == "iso-surface":
                del self._get_api_handle().iso_surface[self._surface_name_on_server]
            elif self.obj.definition.type() == "plane-surface":
                del self._get_api_handle().plane_surface[self._surface_name_on_server]

    def __init__(self, obj):
        """__init__ method of PostAPIHelper class."""
        self.obj = obj
        self.field_info = lambda: obj.get_root().session.fields.field_info
        self.field_data = lambda: obj.get_root().session.fields.field_data
        self.id = lambda: obj.get_root().session.id
        if obj.__class__.__name__ == "Surface":
            self.surface_api = PostAPIHelper._SurfaceAPI(obj)

    @property
    def monitors(self):
        """Returns the session monitors."""
        return self.obj.get_root().session.monitors

    def remote_surface_name(self, local_surface_name):
        """Returns the surface name."""

        local_surfaces_provider = self.obj.get_root()._local_surfaces_provider()

        if local_surface_name in list(local_surfaces_provider):
            return PostAPIHelper._SurfaceAPI.surface_name_on_server(local_surface_name)
        else:
            return local_surface_name

    # Following functions will be deprecated in future.
    def get_vector_fields(self):
        """Returns vector field."""
        return self.field_info().get_vector_fields_info()

    def get_field_unit(self, field):
        """Returns the unit of the field."""
        session = self.obj.get_root().session
        if FluentVersion(session.scheme_eval.version) < FluentVersion.v252:
            quantity = self._field_unit_quantity(field)
            if quantity == "*null*":
                return ""
            scheme_eval_str = f"(units/get-pretty-wb-units-from-dimension (units/inquire-dimension '{quantity}))"
            return " ".join(self._scheme_str_to_py_list(scheme_eval_str))
        else:
            fields_info = self.field_info().get_scalar_fields_info()
            return get_si_unit_for_fluent_quantity(fields_info[field]["quantity_name"])

    def _field_unit_quantity(self, field):
        scheme_eval_str = f"(cdr (assq 'units (%fill-render-info '{field})))"
        return self._scheme_str_to_py_list(scheme_eval_str)[0]

    def _scheme_str_to_py_list(self, scheme_eval_str):
        session = self.obj.get_root().session
        if hasattr(session, "scheme_eval"):
            str_val = session.scheme_eval.string_eval(scheme_eval_str)
            return list(filter(None, re.split(r'[\s()"\']', str_val)))
        else:
            return ["*null*"]
